from __future__ import annotations

from pathlib import Path

import numpy as np
import pytest
from iris import load
from iris.cube import Cube, CubeList
from iris.exceptions import TranslationError

from pyaerocom.exceptions import (
    FileConventionError,
    NetcdfError,
    TemporalResolutionError,
    UnresolvableTimeDefinitionError,
)
from pyaerocom.io import FileConventionRead, iris_io
from tests.fixtures.aeroval import make_dummy_cube_3D_daily
from tests.fixtures.data_access import DataForTests
from tests.fixtures.tm5 import CHECK_PATHS

TM5_FILE = DataForTests(CHECK_PATHS.tm5aod).path


@pytest.fixture()
def cube(defect: Cube | str | None) -> Cube:
    """TM5 AOD cube, with optional deffects"""
    cube: Cube = load(str(TM5_FILE))[0]

    if defect == "notime":
        cube.remove_coord("time")

    if defect == "wrongtime":
        time = cube.coord("time")
        pts = list(time._values_dm._real_array)
        pts.append(70000)
        time._values_dm._real_array = np.asarray(pts)

    if defect == "wrongtime2":
        time = cube.coord("time")
        pts = list(time._values_dm._real_array)
        pts[-1] = 70000
        time._values_dm._real_array = np.asarray(pts)

    if defect == "wrongtime3":
        cube.remove_coord("time")

    if defect == "only_longname_dims":
        cube.coord("longitude").standard_name = None
        cube.coord("longitude").var_name = None
        cube.coord("longitude").long_name = "lon"

        cube.coord("latitude").standard_name = None
        cube.coord("latitude").var_name = None
        cube.coord("latitude").long_name = "lat"

        cube.coord("time").standard_name = None
        cube.coord("time").var_name = None
        cube.coord("time").long_name = "time"

    if defect == "nounit":
        cube.units = ""

    return cube


@pytest.fixture(scope="session")
def fake_file(tmp_path_factory) -> Path:
    """empty/invalid NetCDF file"""
    path: Path = tmp_path_factory.mktemp("test_iris_io") / "invalid.nc"
    path.write_bytes(b"")
    return path


@pytest.fixture()
def file_path(file: str, path_emep: dict[str, Path], fake_file: Path) -> Path:
    """dispatch TM5/EMEP/Empty file path"""
    files = dict(tm5=TM5_FILE, emep=path_emep["monthly"], empty=fake_file)
    try:
        return files[file.casefold()]
    except KeyError:
        raise ValueError(f"Unknown {file=}") from None


@pytest.fixture()
def file_paths(file_path: Path) -> list[Path]:
    """dispatch TM5/EMEP/Empty file path"""
    return [file_path]


@pytest.mark.parametrize("defect", [None])
def test_check_time_coord(cube: Cube):
    iris_io.check_time_coord(cube, "monthly", 2010)


@pytest.mark.parametrize(
    "defect,ts_type,year,exception,error",
    [
        pytest.param(
            "notime",
            "monthly",
            2010,
            AttributeError,
            "Cube does not contain time dimension",
            id="no time dimension",
        ),
        pytest.param(
            None,
            "blaa",
            2010,
            TemporalResolutionError,
            "Invalid input for ts_type blaa. Choose from ['minutely', 'hourly', 'daily', 'weekly', 'monthly', 'yearly', 'native', 'coarsest']",
            id="wrong ts_type",
        ),
        pytest.param(
            None,
            "daily",
            2010,
            UnresolvableTimeDefinitionError,
            "Expected 365 timestamps but data has 12",
            id="daily from monthly",
        ),
        pytest.param(
            None,
            "daily",
            2012,
            UnresolvableTimeDefinitionError,
            "Expected 366 timestamps but data has 12",
            id="daily from monthly",
        ),
        pytest.param(
            None,
            "monthly",
            2008,
            ValueError,
            "First timestamp of data 2010-01-15T12:00:00.000000 does not lie in first period: 2008-01",
            id="worng year",
        ),
    ],
)
def test_check_time_coord_error(
    cube: Cube, ts_type: str, year: int, exception: type[Exception], error: str
):
    with pytest.raises(exception) as e:
        iris_io.check_time_coord(cube, ts_type, year)
    assert str(e.value) == error


@pytest.mark.parametrize("defect", [None])
def test_get_dim_names_cube(cube: Cube):
    assert iris_io.get_dim_names_cube(cube) == ["time", "latitude", "longitude"]


@pytest.mark.parametrize("defect", [None])
def test_get_dimnames_cube(cube: Cube):
    assert iris_io.get_coord_names_cube(cube) == ["time", "latitude", "longitude"]


@pytest.mark.parametrize(
    "defect,file_convention,coords",
    [
        pytest.param(
            None,
            None,
            ["time", "latitude", "longitude"],
            id="TM5",
        ),
        pytest.param(
            None,
            FileConventionRead("aerocom2"),
            ["time", "latitude", "longitude"],
            id="TM5 AeroCom2",
        ),
        pytest.param(
            "notime",
            None,
            ["time", "latitude", "longitude"],
            id="NoTime",
        ),
        pytest.param(
            "notime",
            FileConventionRead("aerocom2"),
            ["latitude", "longitude"],  # no time can be infered under aerocom2
            id="NoTime AeroCom2",
        ),
        pytest.param(
            "wrongtime2",
            None,
            ["time", "latitude", "longitude"],
            id="WrongTime",
        ),
        pytest.param(
            "nounit",
            None,
            ["time", "latitude", "longitude"],
            id="WrongUnit",
        ),
    ],
)
def test__cube_quality_check(cube: Cube, file_convention, coords: list[str]):
    cube = iris_io._cube_quality_check(cube, TM5_FILE.name, file_convention)
    assert [c.name() for c in cube.dim_coords] == coords


@pytest.mark.parametrize("defect", ["wrongtime"])
def test__cube_quality_check_error(cube: Cube):
    with pytest.raises(UnresolvableTimeDefinitionError) as e:
        iris_io._cube_quality_check(cube, TM5_FILE.name)
    error = "UnresolvableTimeDefinitionError('Expected 12 timestamps but data has 13')"
    assert str(e.value) == error


@pytest.mark.parametrize(
    "file,var_name",
    [
        ("tm5", None),
        ("emep", "SURF_ug_NO2"),
    ],
)
def test_load_cube_custom(file_path: Path, var_name: str | None):
    cube = iris_io.load_cube_custom(file_path, var_name)
    assert isinstance(cube, Cube)


@pytest.mark.parametrize(
    "file,var_name,exception,error",
    [
        pytest.param(
            "empty",
            None,
            TranslationError,
            "The file appears empty or incomplete",
            id="enpty file",
        ),
        pytest.param(
            "EMEP",
            None,
            NetcdfError,
            "Could not load single cube from",
            id="no cube",
        ),
        pytest.param(
            "EMEP",
            "od550aer",
            NetcdfError,
            "Variable od550aer not available in file",
            id="no variable",
        ),
    ],
)
def test_load_cube_custom_error(
    file_path: Path, var_name: str | None, exception: type[Exception], error: str
):
    with pytest.raises(exception) as e:
        iris_io.load_cube_custom(file_path, var_name)
    assert str(e.value).startswith(error)


@pytest.mark.parametrize(
    "file,num_loaded",
    [
        ("tm5", 1),
        ("emep", 0),
    ],
)
def test_load_cubes_custom(file_paths: list[Path], num_loaded: int):
    result = iris_io.load_cubes_custom(file_paths)
    assert isinstance(result, tuple) and len(result) == 2
    assert all(isinstance(res, list) for res in result)
    assert all(len(res) == num_loaded for res in result)


@pytest.mark.parametrize("defect", ["only_longname_dims"])
def test_check_dim_coord_names_cube(cube: Cube):
    iris_io.check_dim_coord_names_cube(cube)


@pytest.mark.parametrize("defect", ["wrongtime3", None])
def test__check_correct_time_dim(cube: Cube):
    cube = iris_io._check_correct_time_dim(cube, TM5_FILE.name)
    assert isinstance(cube, Cube)


@pytest.mark.parametrize(
    "defect,file,error",
    [
        pytest.param(
            None,
            "aerocom3_TM5_AP3-CTRL2016_od550aer_Column_-1_monthly.nc",
            "Invalid year -1 in filename aerocom3_TM5_AP3-CTRL2016_od550aer_Column_-1_monthly.nc",
            id="year=-1",
        ),
        pytest.param(
            None,
            "aerocom3_TM5_AP3-CTRL2016_od550aer_Column_20001_monthly.nc",
            "Invalid year 20001 in filename aerocom3_TM5_AP3-CTRL2016_od550aer_Column_20001_monthly.nc",
            id="year=20001",
        ),
    ],
)
def test__check_correct_time_dim_error(cube: Cube, file: str, error: str):
    with pytest.raises(FileConventionError) as e:
        iris_io._check_correct_time_dim(cube, file)
    assert str(e.value) == error


def make_cubelist(dtype):
    return CubeList(
        [
            make_dummy_cube_3D_daily(2010, dtype=dtype),
            make_dummy_cube_3D_daily(2011, dtype=dtype),
        ]
    )


@pytest.mark.parametrize(
    "cubes,val",
    [
        (CubeList([Cube([1]), Cube([1])]), False),
        (CubeList([make_cubelist(int)[0], make_cubelist(float)[0]]), True),
        (make_cubelist(float), False),
        (make_cubelist(int), False),
    ],
)
def test__check_correct_dtypes_timedim_cube_list(cubes, val):
    result = iris_io._check_correct_dtypes_timedim_cube_list(cubes)
    assert result == val


@pytest.mark.parametrize(
    "cubes,sh",
    [
        (make_cubelist(int), (730, 12, 4)),  # see https://github.com/metno/pyaerocom/issues/432
        (make_cubelist(float), (730, 12, 4)),
    ],
)
def test_concatenate_iris_cubes(cubes, sh):
    result = iris_io.concatenate_iris_cubes(cubes)

    assert isinstance(result, Cube)
    assert result.shape == sh



"""
checks if testdata-minimal is available and if not, tries to download it
automatically into ~/MyPyaerocom/testdata-minimal
"""

from __future__ import annotations

import logging
from pathlib import Path
from typing import NamedTuple


from pyaerocom import const, io
from pyaerocom.io.ghost.reader import ReadGhost
from pyaerocom.io.icpforests.reader import ReadICPForest
from pyaerocom.io.readungriddedbase import ReadUngriddedBase

logger = logging.getLogger(__name__)

from pyaerocom import download_minimal_dataset
from pyaerocom.sample_data_access.minimal_dataset import minimal_dataset


class DataForTests(NamedTuple):
    relpath: str
    reader: type[ReadUngriddedBase] | None = None

    @property
    def path(self) -> Path:
        return minimal_dataset.path / "testdata-minimal" / self.relpath

    def register_ungridded(self, name: str):
        if self.reader is None:
            logger.info(f"Adding data search directory {self.path}")
            const.add_data_search_dir(str(self.path))
            return

        if const.OBSLOCS_UNGRIDDED.get(name) == str(self.path):
            logger.info(f"Ungridded dataset {name} is already registered")
            return

        logger.info(f"Register ungridded dataset {name}:  path={self.path} reader={self.reader}")
        const.add_ungridded_obs(name, str(self.path), reader=self.reader, check_read=False)


TEST_DATA: dict[str, DataForTests] = {
    "MODELS": DataForTests("modeldata"),
    "OBSERVATIONS": DataForTests("obsdata"),
    "CONFIG": DataForTests("config"),
    "AeronetSunV3L2Subset.daily": DataForTests(
        "obsdata/AeronetSunV3Lev2.daily/renamed", io.ReadAeronetSunV3
    ),
    "AeronetSunV3L2Subset.AP": DataForTests(
        "obsdata/AeronetSunV3Lev2.0.AP/renamed", io.ReadAeronetSunV3
    ),
    "AeronetSDAV3L2Subset.daily": DataForTests(
        "obsdata/AeronetSDAV3Lev2.daily/renamed", io.ReadAeronetSdaV3
    ),
    "AeronetInvV3L2Subset.daily": DataForTests(
        "obsdata/AeronetInvV3Lev2.daily/renamed", io.ReadAeronetInvV3
    ),
    "EBASSubset": DataForTests("obsdata/EBASMultiColumn", io.ReadEbas),
    "AirNowSubset": DataForTests("obsdata/AirNowSubset", io.ReadAirNow),
    "G.EEA.daily.Subset": DataForTests("obsdata/GHOST/data/EEA_AQ_eReporting/daily", ReadGhost),
    "G.EEA.hourly.Subset": DataForTests("obsdata/GHOST/data/EEA_AQ_eReporting/hourly", ReadGhost),
    "G.EBAS.daily.Subset": DataForTests("obsdata/GHOST/data/EBAS/daily", ReadGhost),
    "G.EBAS.hourly.Subset": DataForTests("obsdata/GHOST/data/EBAS/hourly", ReadGhost),
    "ICPFORESTS.Subset": DataForTests("obsdata/ipc-forests/dep", ReadICPForest),
    "EEA_AQeRep.v2.Subset": DataForTests("obsdata/EEA_AQeRep.v2/renamed", io.ReadEEAAQEREP_V2),
    "Earlinet-test": DataForTests("obsdata/Earlinet", io.ReadEarlinet),
}


def init() -> bool:
    download_minimal_dataset()

    for name, data in TEST_DATA.items():
        assert data.path.is_dir(), f"missing dataset {name=}"
        try:
            data.register_ungridded(name)
        except Exception as e:
            logger.warning(
                f"Failed to instantiate testdata since ungridded "
                f"dataset {name} at {data.path} could not be registered: {e}"
            )
            return False

    return True


assert init(), "cound not find minimal test data"



import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt

from pyaerocom._lowlevel_helpers import BrowseDict


class VerticalProfile:
    """Object representing single variable profile data"""

    def __init__(
        self,
        data: npt.ArrayLike,
        altitude: npt.ArrayLike,
        dtime,
        var_name: str,
        data_err: npt.ArrayLike | None,
        var_unit: str,
        altitude_unit: str,
    ):
        self.var_name = var_name
        self.dtime = dtime
        self.data = data
        self.data_err = data_err
        self.altitude = altitude

        self.var_info = BrowseDict()
        self.var_info["altitude"] = dict(units=altitude_unit)
        self.var_info[self.var_name] = dict(units=var_unit)

        # Guard against having data (and data errors) with missing asociated altitude info
        if hasattr(self.data_err, "__len__"):
            assert len(self.data) == len(self.data_err) == len(self.altitude)
        else:
            assert len(self.data) == len(self.altitude)

    @property
    def data(self):
        """Array containing data values corresponding to data"""
        return self._data

    @data.setter
    def data(self, val):
        if isinstance(val, list):
            val = np.asarray(val)
        self._data = val

    @property
    def data_err(self):
        """Array containing data values corresponding to data"""
        return self._data_err

    @data_err.setter
    def data_err(self, val):
        if isinstance(val, list):
            val = np.asarray(val)
        self._data_err = val

    @property
    def altitude(self):
        """Array containing altitude values corresponding to data"""
        return self._altitude

    @altitude.setter
    def altitude(self, val):
        if isinstance(val, list):
            val = np.asarray(val)
        self._altitude = val

    # ToDo: complete docstring
    def plot(
        self,
        plot_errs=True,
        whole_alt_range=False,
        rot_xlabels=30,
        errs_shaded=True,
        errs_alpha=0.1,
        add_vertbar_zero=True,
        figsize=None,
        ax=None,
        **kwargs,
    ):  # pragma: no cover
        """Simple plot method for vertical profile"""
        if figsize is None:
            figsize = (4, 8)
        if ax is None:
            _, ax = plt.subplots(1, 1, figsize=figsize)

        p = ax.plot(self.data, self.altitude, "-x", **kwargs)

        c = p[0].get_color()
        if rot_xlabels:
            for lbl in ax.get_xticklabels():
                lbl.set_rotation(rot_xlabels)

        unit = self.var_info[self.var_name]["units"]
        aunit = self.var_info["altitude"]["units"]

        xlab = f"{self.var_name} [{unit}]"
        ylab = f"Altitude [{aunit}]"

        ax.set_xlabel(xlab)
        ax.set_ylabel(ylab)

        if whole_alt_range:
            ax.set_ylim([np.min([0, self.altitude.min()]), self.altitude.max()])
        if plot_errs:
            lower = self.data - self.data_err
            upper = self.data + self.data_err
            if errs_shaded:
                ax.fill_betweenx(self.altitude, lower, upper, color=c, alpha=errs_alpha)
            else:
                ax.errorbar(
                    self.data,
                    self.altitude,
                    xerr=self.data_err,
                    ls=" ",
                    marker=" ",
                    color="#cccccc",
                )
        if add_vertbar_zero:
            xl = ax.get_xlim()
            if xl[0] < 0 < xl[1]:
                ax.plot([0, 0], ax.get_ylim(), "--", color="#cccccc")
        ax.figure.tight_layout()

        return ax



from __future__ import annotations

from pathlib import Path

import pytest
from typer.testing import CliRunner

from pyaerocom.scripts.cams2_83.cli import app

runner = CliRunner()


@pytest.mark.usefixtures("fake_ExperimentProcessor", "reset_cachedir")
def test_clearcache(
    monkeypatch,
    fake_cache_path: Path,
    tmp_path: Path,
    caplog,
):
    assert list(fake_cache_path.glob("*.pkl"))

    options = f"forecast week 2024-03-16 2024-03-23 --model-path {tmp_path} --obs-path {tmp_path} --data-path {tmp_path} --coldata-path {tmp_path} --cache {fake_cache_path} --name 'Test'"
    result = runner.invoke(app, options.split())
    assert "Running Statistics" in caplog.text
    assert result.exit_code == 0
    # Check that the cache is cleared
    assert not list(fake_cache_path.glob("*.pkl"))


@pytest.mark.usefixtures("fake_CAMS2_83_Processer", "reset_cachedir")
def test_not_cleared_cache(
    monkeypatch,
    fake_cache_path: Path,
    tmp_path: Path,
    caplog,
):
    assert list(fake_cache_path.glob("*.pkl"))

    options = f"forecast long 2024-03-16 2024-03-23 --model-path {tmp_path} --obs-path {tmp_path} --data-path {tmp_path} --coldata-path {tmp_path} --name 'Test' --medianscores"
    result = runner.invoke(app, options.split())
    assert "Running CAMS2_83 Specific Statistics, cache is not cleared" in caplog.text
    assert result.exit_code == 0
    # Check that the cache is not cleared
    assert list(fake_cache_path.glob("*.pkl"))


def test_eval_dummy(
    tmp_path: Path,
    caplog,
):
    options = f"forecast day 2024-03-16 2024-03-16 --model-path {tmp_path} --obs-path {tmp_path} --data-path {tmp_path} --coldata-path {tmp_path} --name 'Test'"
    result = runner.invoke(app, options.split())
    assert result.exit_code == 0
    assert "Failed to read model variable" in caplog.text


def test_eval_medianscores_dummy(
    tmp_path: Path,
    caplog,
):
    options = f"analysis long 2023-03-01 2024-02-28 --model-path {tmp_path} --obs-path {tmp_path} --data-path {tmp_path} --coldata-path {tmp_path} --name 'Test' --medianscores"
    result = runner.invoke(app, options.split())
    assert result.exit_code == 0
    assert "Running CAMS2_83 Specific Statistics, cache is not cleared" in caplog.text
    assert "Failed to read model variable" in caplog.text



from pathlib import Path

import pytest

from pyaerocom.exceptions import AeronetReadError
from pyaerocom.io.read_aeronet_sunv3 import ReadAeronetSunV3


@pytest.fixture(scope="module")
def reader():
    return ReadAeronetSunV3("AeronetSunV3L2Subset.AP")


def test_get_file_list(reader):
    assert len(reader.get_file_list()) >= 2


def test_read_file(reader):
    from pyaerocom.stationdata import StationData

    reader.get_file_list()
    file = reader.files[0]
    assert Path(file).name == "19930101_20211120_Karlsruhe.lev20.gz"
    data = reader.read_file(file)
    assert isinstance(data, StationData)
    assert data.latitude[0] == pytest.approx(49.09, 0.01)
    assert data.longitude[0] == pytest.approx(8.42, 0.01)
    assert data.station_name[0] == "Karlsruhe"
    assert all(x in data for x in ["od550aer", "ang4487aer"])


def test_read_add_common_meta(reader):
    reader.get_file_list()
    files = reader.files[0]

    data = reader.read("od550aer", files=files, common_meta={"bla": 42})
    assert all("bla" in x for x in data.metadata.values())


def test_exception_aeronet_read_error(reader):
    reader.get_file_list()
    files = sorted(reader.files)[-1]
    with pytest.raises(AeronetReadError) as e:
        data = reader.read_file(files[-1])
        assert data
    assert str(e.value).startswith("gzip error in file")



from __future__ import annotations

import pytest

from pyaerocom.aeroval.experiment_output import ExperimentOutput
from pyaerocom.aeroval.experiment_processor import ExperimentProcessor
from pyaerocom.aeroval import EvalSetup
from tests.conftest import geojson_unavail


@pytest.mark.parametrize("cfg", ["cfgexp1"])
def test_ExperimentProcessor___init__(eval_config: dict):
    cfg = EvalSetup(**eval_config)
    proc = ExperimentProcessor(cfg)
    assert isinstance(proc.cfg, EvalSetup)
    assert isinstance(proc.exp_output, ExperimentOutput)


@pytest.fixture
def processor(eval_config: dict) -> ExperimentProcessor:
    """ExperimentProcessor instance without experiment data"""
    setup = EvalSetup(**eval_config)
    proc = ExperimentProcessor(setup)
    proc.exp_output.delete_experiment_data(also_coldata=True)
    return proc


@geojson_unavail
@pytest.mark.parametrize("cfg", ["cfgexp1", "cfgexp2", "cfgexp3", "cfgexp4", "cfgexp5"])
def test_ExperimentProcessor_run(processor: ExperimentProcessor):
    processor.run()


# Temporary until ObsCollection implemented simiarly then can run same test
@geojson_unavail
@pytest.mark.parametrize(
    "cfg,kwargs,error",
    [
        (
            "cfgexp2",
            dict(model_name="BLA"),
            "'No matches could be found that match input BLA'",
        ),
        (
            "cfgexp2",
            dict(obs_name="BLUB"),
            "'No matches could be found that match input BLUB'",
        ),
    ],
)
def test_ExperimentProcessor_run_error_obs_name(
    processor: ExperimentProcessor, kwargs: dict, error: str
):
    with pytest.raises(KeyError) as e:
        processor.run(**kwargs)
    assert str(e.value) == error


@geojson_unavail
@pytest.mark.parametrize(
    "cfg,kwargs,error",
    [
        ("cfgexp2", dict(var_list="BLUB"), "var_list %s and obs_vars %s mismatch."),
    ],
)
def test_ExperimentProcessor_catch_wrong_var_list(
    processor: ExperimentProcessor, kwargs: dict, error: str, caplog
):
    processor.run(**kwargs)
    assert any([error in str(record) for record in caplog.records])



from __future__ import annotations

import logging
from typing import TypeVar, Generic
import functools
from collections import defaultdict

import numpy as np
import numpy.typing as npt
from pyaro import list_timeseries_engines, open_timeseries
from pyaro.timeseries import Data, Reader, Station
from pyaro.timeseries.Wrappers import VariableNameChangingReader

from pyaerocom.io.pyaro.pyaro_config import PyaroConfig
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
from pyaerocom.tstype import TsType
from pyaerocom.ungriddeddata import UngriddedData

logger = logging.getLogger(__name__)


class ReadPyaro(ReadUngriddedBase):
    __version__ = "1.1.0"

    SUPPORTED_DATASETS = list(list_timeseries_engines().keys())

    def __init__(self, config: PyaroConfig) -> None:
        self.config: PyaroConfig = config

        self._check_id()

        self.converter = PyaroToUngriddedData(self.config)
        self.reader = self.converter.reader
        self._data_dir = self.config.filename_or_obj_or_url
        self._data_name = self.config.name
        self._data_id = self.config.name

    """
    Definition of abstract methods from ReadUngriddedBase
    """

    @property
    def DATA_ID(self):
        return self._data_name

    @property
    def PROVIDES_VARIABLES(self):
        """
        return self.reader.get_variables()
        """
        return self.reader.variables()

    @property
    def DEFAULT_VARS(self):
        return self.PROVIDES_VARIABLES

    @property
    def TS_TYPE(self):
        """
        To be provided by the reader or engine
        """
        return "undefined"

    @property
    def _FILEMASK(self):
        return self.config.filename_or_obj_or_url

    @staticmethod
    def get_pyaro_readers():
        return list_timeseries_engines()

    def read(self, vars_to_retrieve=None, files=..., first_file=None, last_file=None):
        return self.converter.read(vars_to_retrieve=vars_to_retrieve)

    def read_file(self, filename, vars_to_retrieve=None):
        return self.converter.read(vars_to_retrieve)

    def _check_id(self):
        avail_readers = list_timeseries_engines()
        if self.config.reader_id not in avail_readers:
            logger.warning(
                f"Could not find {self.config.reader_id} in list of available Pyaro readers: {avail_readers}"
            )


def _calculate_ts_type(
    start: npt.NDArray[np.datetime64], end: npt.NDArray[np.datetime64]
) -> npt.NDArray[TsType]:
    seconds = (end - start).astype("timedelta64[s]").astype(np.int32)

    @np.vectorize(otypes=[TsType])
    @functools.lru_cache(maxsize=128)
    def memoized_ts_type(x: np.int32) -> TsType:
        if x == 0:
            return TsType("hourly")
        return TsType.from_total_seconds(x)

    return memoized_ts_type(seconds)


class PyaroToUngriddedData:
    def __init__(self, config: PyaroConfig) -> None:
        self.data: UngriddedData = UngriddedData()
        self.config = config
        self.reader: Reader = self._open_reader()

    def _open_reader(self) -> Reader:
        reader_id = self.config.reader_id
        if self.config.model_extra is not None:
            kwargs = self.config.model_extra
        else:
            kwargs = {}

        if self.config.name_map is None:
            return open_timeseries(
                reader_id,
                self.config.filename_or_obj_or_url,
                filters=self.config.filters,
                **kwargs,
            )
        else:
            return VariableNameChangingReader(
                open_timeseries(
                    reader_id,
                    self.config.filename_or_obj_or_url,
                    filters=self.config.filters,
                    **kwargs,
                ),
                self.config.name_map,
            )

    def _convert_to_ungriddeddata(self, pyaro_data: dict[str, Data]) -> UngriddedData:
        total_size = sum(len(var) for var in pyaro_data.values())

        COLNO = 12
        outarray = np.nan * np.ones((total_size, COLNO), dtype=float, order="F")

        T = TypeVar("T")

        class UniqueMapper(Generic[T]):
            """Assign a unique ID to each item it has
            not seen before
            """

            def __init__(self):
                self.inner: dict[T, float] = dict()
                self._idx: float = 0.0

            def index_and_insert_if_not(self, item: T) -> float:
                index = self.inner.get(item)
                if index is not None:
                    return index
                index = self._idx
                self._idx += 1.0

                self.inner[item] = index
                return index

            def __getitem__(self, key: T) -> float:
                return self.inner[key]

        # unique in (station_name, variable_name, units, tstype)
        station_mapper: UniqueMapper[tuple[str, str, str, str]] = UniqueMapper()
        var_mapper: UniqueMapper[str] = UniqueMapper()

        stations_with_metadata = self.get_stations()

        current_offset = 0
        for var, var_data in pyaro_data.items():
            next_offset = current_offset + len(var_data)
            idx = slice(current_offset, next_offset)
            current_offset = next_offset

            var_key = var_mapper.index_and_insert_if_not(var)

            tstype = _calculate_ts_type(var_data.start_times, var_data.end_times)
            stations = var_data.stations
            units = var_data.units
            # Find unique pairs of (station, var, TsType)
            station_key = [
                station_mapper.index_and_insert_if_not((s, var, units, str(t)))
                for (s, t) in zip(stations, tstype)
            ]

            outarray[idx, UngriddedData._METADATAKEYINDEX] = station_key
            # midtime = var_data.start_times + (var_data.end_times - var_data.start_times)/2
            # outarray[idx, UngriddedData._TIMEINDEX] = midtime.astype("datetime64[s]")
            outarray[idx, UngriddedData._TIMEINDEX] = var_data.start_times.astype("datetime64[s]")
            outarray[idx, UngriddedData._LATINDEX] = var_data.latitudes
            outarray[idx, UngriddedData._LONINDEX] = var_data.longitudes
            outarray[idx, UngriddedData._ALTITUDEINDEX] = var_data.altitudes
            outarray[idx, UngriddedData._VARINDEX] = var_key
            outarray[idx, UngriddedData._DATAINDEX] = var_data.values
            # outarray[idx, UngriddedData._DATAHEIGHTINDEX] = ?? Unused ??
            outarray[idx, UngriddedData._DATAERRINDEX] = var_data.standard_deviations
            outarray[idx, UngriddedData._DATAFLAGINDEX] = var_data.flags  # Only counts if non-NaN?
            # outarray[idx, UngriddedData._STOPTIMEINDEX] = var_data.end_times # Seems unused?
            # outarray[idx, UngriddedData._TRASHINDEX]  # No need to set, only non-NaN values are considered trash

        metadata = dict()
        for (station_name, var, units, tstype), station_key in station_mapper.inner.items():
            extra_metadata = stations_with_metadata[station_name].metadata
            d = {
                "data_id": self.config.name,
                "station_name": station_name,
                "var_info": {
                    var: {"units": units},
                },
                **stations_with_metadata[station_name],
                **extra_metadata,
            }
            if "ts_type" not in d:
                d["ts_type"] = tstype
            metadata[station_key] = d

        meta_idx = defaultdict(dict)
        for (_station_name, var, _units, tstype), station_key in station_mapper.inner.items():
            var_key = var_mapper[var]
            mask = (outarray[:, UngriddedData._METADATAKEYINDEX] == station_key) & (
                outarray[:, UngriddedData._VARINDEX] == var_key
            )
            indices = np.flatnonzero(mask)
            meta_idx[station_key][var] = indices

        var_idx = var_mapper.inner

        return UngriddedData._from_raw_parts(outarray, metadata, meta_idx, var_idx)

    def get_variables(self) -> list[str]:
        return self.reader.variables()

    def get_stations(self) -> dict[str, Station]:
        return self.reader.stations()

    def read(self, vars_to_retrieve=None) -> UngriddedData:
        allowed_vars = self.get_variables()
        if vars_to_retrieve is None:
            vars_to_retrieve = allowed_vars
        else:
            if isinstance(vars_to_retrieve, str):
                vars_to_retrieve = [vars_to_retrieve]

        data = {}
        for var in vars_to_retrieve:
            if var not in allowed_vars:
                logger.warning(
                    f"Variable {var} not in list over allowed variabes for {self.config.reader_id}: {allowed_vars}"
                )
                continue

            data[var] = self.reader.data(varname=var)

        return self._convert_to_ungriddeddata(data)



import abc
import logging

from pyaerocom._lowlevel_helpers import TypeValidator
from pyaerocom.aeroval import EvalSetup
from pyaerocom.aeroval.experiment_output import ExperimentOutput
from pyaerocom.colocation.colocation_setup import ColocationSetup
from pyaerocom.colocation.colocator import Colocator

logger = logging.getLogger(__name__)


class HasConfig:
    """
    Base class that ensures that evaluation configuration is available

    Attributes
    ----------
    cfg : EvalSetup
        AeroVal experiment setup
    exp_output : ExperimentOutput
        Manages output for an AeroVal experiment (e.g. path locations).

    """

    cfg = TypeValidator(EvalSetup)
    exp_output = TypeValidator(ExperimentOutput)

    def __init__(self, cfg: EvalSetup):
        self.cfg = cfg
        self.exp_output = ExperimentOutput(cfg)
        self.avdb = self.exp_output.avdb

    @property
    def raise_exceptions(self):
        return self.cfg.colocation_opts.raise_exceptions

    @property
    def reanalyse_existing(self):
        return self.cfg.colocation_opts.reanalyse_existing


class ProcessingEngine(HasConfig, abc.ABC):
    """
    Abstract base for classes supposed to do one or more processing tasks

    Requirement for a processing class is to inherit attrs from
    :class:`HasConfig` and, in addition to that, to have implemented a method
    :fun:`run` which is running the corresponding processing task and storing
    all the associated output files, that are read by the frontend.

    One example of an implementation is the
    :class:`pyaerocom.aeroval.modelmaps_engine.ModelMapsEngine`.
    """

    @abc.abstractmethod
    def run(self, *args, **kwargs) -> list:
        """
        Method that runs the processing based on settings in :attr:`cfg`

        Parameters
        ----------
        *args
            positional arguments.
        **kwargs
            Keyword arguments.

        Returns
        -------
        list
            list of output file paths generated by the engine.

        """
        pass


class HasColocator(HasConfig):
    """
    Config class that also has the ability to co-locate
    """

    def get_colocator(self, model_name: str = None, obs_name: str = None) -> Colocator:
        """
        Instantiate colocation engine

        Parameters
        ----------
        model_name : str, optional
            name of model. The default is None.
        obs_name : str, optional
            name of obs. The default is None.

        Returns
        -------
        Colocator

        """
        col_cfg = {**self.cfg.colocation_opts.model_dump()}
        outdir = self.cfg.path_manager.get_coldata_dir()
        col_cfg["basedir_coldata"] = outdir

        if not model_name and not obs_name:
            col_stp = ColocationSetup(**col_cfg)
            return Colocator(col_stp)

        if model_name:
            mod_cfg = self.cfg.get_model_entry(model_name)
            col_cfg["model_cfg"] = mod_cfg

            # Hack and at what lowlevel_helpers's import_from was doing
            for key, val in mod_cfg.items():
                if key in ColocationSetup.model_fields:
                    col_cfg[key] = val
        if obs_name:
            obs_cfg = self.cfg.get_obs_entry(obs_name)
            pyaro_config = obs_cfg["obs_config"] if "obs_config" in obs_cfg else None
            col_cfg["obs_config"] = pyaro_config

            # Hack and at what lowlevel_helpers's import_from was doing
            for key, val in obs_cfg.model_dump().items():
                if key in ColocationSetup.model_fields:
                    col_cfg[key] = val

            col_cfg["add_meta"].update(diurnal_only=self.cfg.get_obs_entry(obs_name).diurnal_only)

        col_stp = ColocationSetup(**col_cfg)
        col = Colocator(col_stp)

        return col


class DataImporter(HasColocator):
    """
    Class that supports reading of model and obs data based on an eval config.

    Depending on a :class:`EvalSetup`, reading of model and obs data may have
    certain constraints (e.g. freq, years, alias variable names, etc.), which
    are / can be specified flexibly for each model and obs entry in an
    analysis setup (:class:`EvalSetup`). Proper handling of these reading
    constraints and data import settings are handled in the
    :class:`pyaerocom.colocation.Colocator` engine, therefore the reading
    in this class is done via the :class:`Colocator` engine.


    """

    def read_model_data(self, model_name, var_name):
        """
        Import model data

        Parameters
        ----------
        model_name : str
            Name of model in :attr:`cfg`,
        var_name : str
            Name of variable to be read.

        Returns
        -------
        data : GriddedData
            loaded model data.

        """
        col = self.get_colocator(model_name=model_name)
        data = col.get_model_data(var_name)

        return data

    def read_ungridded_obsdata(self, obs_name, var_name):
        """
        Import ungridded observation data

        Parameters
        ----------
        obs_name : str
            Name of observation network in :attr:`cfg`
        var_name : str
            Name of variable to be read.

        Returns
        -------
        data : UngriddedData
            loaded obs data.

        """

        col = self.get_colocator(obs_name=obs_name)

        data = col._read_ungridded(var_name)
        return data

    def read_gridded_obsdata(self, obs_name, var_name):
        """
        Import gridded observation data, usually satellite data

        Args:
            obs_name (str): Name of observation network in :attr:`cfg`
            var_name (str): Name of variable to be read.
        """
        col = self.get_colocator(obs_name=obs_name)
        data = col._read_gridded(var_name, is_model=False)
        return data



"""
This module contains functionality related to regions in pyaerocom
"""

from __future__ import annotations

import numpy as np

from pyaerocom._lowlevel_helpers import BrowseDict
from pyaerocom.config import ALL_REGION_NAME
from pyaerocom.helpers_landsea_masks import get_mask_value, load_region_mask_xr
from pyaerocom.region_defs import HTAP_REGIONS  # list of HTAP regions
from pyaerocom.region_defs import REGION_DEFS  # all region definitions
from pyaerocom.region_defs import OLD_AEROCOM_REGIONS, REGION_NAMES  # custom names (dict)

POSSIBLE_REGION_OCEAN_NAMES = ["OCN", "Oceans"]


class Region(BrowseDict):
    """Class specifying a region

    Attributes
    ----------
    region_id : str
        ID of region (e.g. EUROPE)
    name : str
        name of region (e.g. Europe) used e.g. in plotting.
    lon_range : list
        longitude range (min, max) covered by region
    lat_range : list
        latitude range (min, max) covered by region
    lon_range_plot : list
        longitude range (min, max) used for plotting region.
    lat_range_plot : list
        latitude range (min, max) used for plotting region.
    lon_ticks : list
        list of longitude ticks used for plotting
    lat_ticks : list
        list of latitude ticks used for plotting

    Parameters
    ----------
    region_id : str
        ID of region (e.g. "EUROPE"). If the input region ID is registered as
        a default region in :mod:`pyaerocom.region_defs`, then the default
        information is automatically imported on class instantiation.
    **kwargs
        additional class attributes (see above for available default attributes).
        Note, any attr. values provided by kwargs are preferred over
        potentially defined default attrs. that are imported automatically.
    """

    def __init__(self, region_id=None, **kwargs):
        if region_id is None:
            region_id = ALL_REGION_NAME

        if region_id in REGION_NAMES:
            name = REGION_NAMES[region_id]
        else:
            name = region_id

        self.region_id = region_id
        self.name = name

        self.lon_range = None
        self.lat_range = None

        # longitude / latitude range of data in plots
        self.lon_range_plot = None
        self.lat_range_plot = None

        self.lon_ticks = None
        self.lat_ticks = None

        self._mask_data = None
        if region_id in REGION_DEFS:
            self.import_default(region_id)

        self.update(**kwargs)

    def is_htap(self):
        """Boolean specifying whether region is an HTAP binary region"""
        return True if self.region_id in HTAP_REGIONS else False

    def import_default(self, region_id):
        """Import region definition

        Parameters
        ----------
        region_id : str
            ID of region

        Raises
        ------
        KeyError
            if no region is registered for the input ID
        """
        self.update(REGION_DEFS[region_id])

        if self.lon_range_plot is None:
            self.lon_range_plot = self.lon_range
        if self.lat_range_plot is None:
            self.lat_range_plot = self.lat_range

    @property
    def center_coordinate(self):
        """Center coordinate of this region"""
        latc = self.lat_range[0] + (self.lat_range[1] - self.lat_range[0]) / 2
        lonc = self.lon_range[0] + (self.lon_range[1] - self.lon_range[0]) / 2
        return (latc, lonc)

    def distance_to_center(self, lat, lon):
        """Compute distance of input coordinate to center of this region

        Parameters
        ----------
        lat : float
            latitude of coordinate
        lon : float
            longitude of coordinate

        Returns
        -------
        float
            distance in km
        """
        from pyaerocom.geodesy import calc_distance

        cc = self.center_coordinate
        return calc_distance(lat0=cc[0], lon0=cc[1], lat1=lat, lon1=lon)

    def contains_coordinate(self, lat, lon):
        """Check if input lat/lon coordinate is contained in region

        Parameters
        ----------
        lat : float
            latitude of coordinate
        lon : float
            longitude of coordinate

        Returns
        -------
        bool
            True if coordinate is contained in this region, False if not
        """

        lat_lb = self.lat_range[0]
        lat_ub = self.lat_range[1]
        lon_lb = self.lon_range[0]
        lon_ub = self.lon_range[1]
        # latitude bounding boxes should always be defined with the southern most boundary less than the northernmost
        lat_ok = lat_lb <= lat <= lat_ub
        # if the longitude bounding box has a lowerbound less than the upperbound
        if lon_lb < lon_ub:
            # it suffices to check that lon is between these values
            lon_ok = lon_lb <= lon <= lon_ub
        # if the longitude lowerbound has a value lessthan the upperbound
        elif lon_ub < lon_lb:
            # lon is contained in the bounding box in two cases
            lon_ok = lon < lon_ub or lon > lon_lb
        else:
            lon_ok = False  # safeguard
        return lat_ok * lon_ok

    def mask_available(self):
        if not self.is_htap():
            return False
        return True

    def get_mask_data(self):
        if not self.mask_available():
            raise AttributeError(f"No binary mask data available for region {self.region_id}.")
        if self._mask_data is None:
            self._mask_data = load_region_mask_xr(self.region_id)
        return self._mask_data

    def plot_mask(self, ax, color, alpha=0.2):
        mask = self.get_mask_data()
        # import numpy as np
        data = mask.data
        data[data == 0] = np.nan
        mask.data = data

        mask.plot(ax=ax)
        return ax

    def plot_borders(self, ax, color, lw=2):
        raise NotImplementedError("Coming soon...")

    def plot(self, ax=None):
        """
        Plot this region

        Draws a rectangle of the outer bounds of the region and if a binary
        mask is available for this region, it will be plotted as well.

        Parameters
        ----------
        ax : GeoAxes, optional
            axes instance to be used for plotting. Defaults to None in which
            case a new instance is created.

        Returns
        -------
        GeoAxes
            axes instance used for plotting

        """
        from cartopy.mpl.geoaxes import GeoAxes

        from pyaerocom.plot.mapping import init_map

        if ax is None:
            ax = init_map()
        elif not isinstance(ax, GeoAxes):
            raise ValueError("Invalid input for ax: need cartopy GeoAxes..")

        if self.mask_available():
            self.plot_mask(ax, color="r")

        ax.set_xlabel("Longitude")
        ax.set_ylabel("Latitude")
        name = self.name
        if not name == self.region_id:
            name += f" (ID={self.region_id})"

        ax.set_title(name)

        return ax

    def __contains__(self, val):
        if not isinstance(val, tuple):
            raise TypeError("Invalid input, need tuple")
        if not len(val) == 2:
            raise ValueError("Invalid input: coordinate must contain 2 elements (lat, lon)")
        return self.contains_coordinate(lat=val[0], lon=val[1])

    def __repr__(self):
        return f"Region {self.name} {super().__repr__()}"

    def __str__(self):
        s = (
            f"pyaeorocom Region\nName: {self.name}\n"
            f"Longitude range: {self.lon_range}\n"
            f"Latitude range: {self.lat_range}\n"
            f"Longitude range (plots): {self.lon_range_plot}\n"
            f"Latitude range (plots): {self.lat_range_plot}"
        )
        return s


def all():
    """Wrapper for :func:`get_all_default_region_ids`"""
    return list(REGION_DEFS)


def get_all_default_region_ids():
    """Get list containing IDs of all default regions

    Returns
    -------
    list
        IDs of all predefined default regions
    """
    return OLD_AEROCOM_REGIONS


def _get_regions_helper(reg_ids):
    """
    Get dictionary of :class:`Region` instances for input IDs

    Parameters
    ----------
    reg_ids : list
        list of region IDs

    Returns
    -------
    dict
        keys are input region IDs, values are loaded :class:`Region` instances
    """
    regs = {}
    for reg in reg_ids:
        regs[reg] = Region(reg)
    return regs


def get_old_aerocom_default_regions():
    """
    Load dictionary with default AeroCom regions

    Returns
    -------
    dict
        keys are region ID's, values are instances of :class:`Region`
    """
    return _get_regions_helper(OLD_AEROCOM_REGIONS)


def get_htap_regions():
    """
    Load dictionary with HTAP regions

    Returns
    -------
    dict
        keys are region ID's, values are instances of :class:`Region`
    """
    return _get_regions_helper(HTAP_REGIONS)


def get_all_default_regions():
    """Get dictionary containing all default regions from region.ini file

    Returns
    -------
    dict
        dictionary containing all default regions; keys are region ID's, values
        are instances of :class:`Region`.

    """
    return get_old_aerocom_default_regions()


#: ToDO: check how to handle methods properly with HTAP regions...
def get_regions_coord(lat, lon, regions=None):
    """Get the region that contains an input coordinate

    Note
    ----
    This does not yet include HTAP, since this causes troules in automated
    AeroCom processing

    Parameters
    ----------
    lat : float
        latitude of coordinate
    lon : float
        longitude of coordinate
    regions : dict, optional
        dictionary containing instances of :class:`Region` as values, which
        are considered. If None, then all default regions are used.

    Returns
    -------
    list
        list of regions that contain this coordinate
    """
    matches = []
    if regions is None:
        regions = get_all_default_regions()
    ocean_mask = load_region_mask_xr("OCN")
    on_ocean = bool(get_mask_value(lat, lon, ocean_mask))
    for rname, reg in regions.items():
        if rname == ALL_REGION_NAME:  # always True for ALL_REGION_NAME
            matches.append(rname)
            continue
        # OCN needs special handling determined by the rname, not hardcoded to return OCN b/c of HTAP issues
        if rname in POSSIBLE_REGION_OCEAN_NAMES:
            if on_ocean:
                matches.append(rname)
            continue
        if reg.contains_coordinate(lat, lon) and not on_ocean:
            matches.append(rname)
    if len(matches) == 0:
        matches.append(ALL_REGION_NAME)
    return matches


def find_closest_region_coord(
    lat: float, lon: float, regions: dict | None = None, **kwargs
) -> list[str]:
    """Finds list of regions sorted by their center closest to input coordinate

    Parameters
    ----------
    lat : float
        latitude of coordinate
    lon : float
        longitude of coordinate
    regions : dict, optional
        dictionary containing instances of :class:`Region` as values, which
        are considered. If None, then all default regions are used.

    Returns
    -------
    list[str]
        sorted list of region IDs of identified regions
    """
    if regions is None:
        regions = get_all_default_regions()
    matches = get_regions_coord(lat, lon, regions)
    matches.sort(key=lambda id: regions[id].distance_to_center(lat, lon))
    if kwargs.get("regions_how") == "htap":
        # keep only first entry and Oceans if it exists
        keep = matches[:1]
        if "Oceans" in matches[1:]:
            keep += ["Oceans"]
        if ALL_REGION_NAME in matches[1:]:
            keep += [ALL_REGION_NAME]
        return list(set(keep))
    return matches



import numpy as np
import pandas as pd
import pytest
from matplotlib.axes import Axes
from matplotlib.colors import BoundaryNorm

from pyaerocom.plot.heatmaps import df_to_heatmap

colnames = ["M1", "M2", "M3"]
rownames = ["O1", "O2", "O3", "O4"]
arr = np.ones((len(rownames), len(colnames)))
arr *= np.asarray([1, 2, 3])

df = pd.DataFrame(arr, index=rownames, columns=colnames)

strannot = df.values.astype(str)
annot_nans = np.nan * df.values
one_nan = df.values
one_nan[0, 0] = np.nan


@pytest.mark.parametrize(
    "kwargs",
    [
        dict(),
        dict(sub_norm_before_div=False, normalise_rows=True),
        dict(normalise_rows=True),
        dict(norm_ref=0.1, normalise_rows=True),
        dict(normalise_rows=True, normalise_rows_how="mean"),
        dict(normalise_rows=True, normalise_rows_how="median"),
        dict(normalise_rows=True, normalise_rows_how="median"),
        dict(normalise_rows=True, normalise_rows_col="M3"),
        dict(color_rowwise=True),
        dict(norm=BoundaryNorm(boundaries=[1, 2, 3], ncolors=10, clip=False)),
        dict(annot=None),
        dict(annot=strannot),
        dict(annot_fmt_rowwise=True),
        dict(annot_fmt_rowwise=True, annot=annot_nans),
        dict(annot=df.values, num_digits=2),
        dict(annot_fmt_rowwise=True, annot_fmt_exceed=[2, ".1E"]),
        dict(annot_fmt_rowwise=True, annot=one_nan),
        dict(title="Blaa"),
        dict(cbar_labelsize=12),
        dict(annot_fmt_rowwise=True, annot=df.values * 1e-6),
        dict(annot_fmt_rowwise=True, annot=df.values * 1e-2),
        dict(annot_fmt_rowwise=True, annot=df.values * 1e6),
    ],
)
@pytest.mark.filterwarnings("ignore:More than 20 figures have been opened:RuntimeWarning")
def test_df_to_heatmap(kwargs):
    val = df_to_heatmap(df, **kwargs)
    assert isinstance(val[0], Axes)


@pytest.mark.parametrize(
    "kwargs,error",
    [
        pytest.param(
            dict(annot=[1, 2, 3]),
            "Invalid input for annot: needs to have same shape as input dataframe",
            id="invalid annot",
        ),
        pytest.param(
            dict(normalise_rows=True, normalise_rows_how="sum"),
            "Invalid input for normalise_rows_how (sum). Choose mean or median",
            id="invalid normalise_rows_how",
        ),
        pytest.param(
            dict(normalise_rows=True, normalise_rows_how="BLAA"),
            "Invalid input for normalise_rows_how (BLAA). Choose mean or median",
            id="invalid normalise_rows_how",
        ),
        pytest.param(
            dict(normalise_rows=True, normalise_rows_col="BLAAA"),
            "Failed to localise column BLAAA",
            id="invalid column",
        ),
    ],
)
def test_df_to_heatmap_error(kwargs, error: str):
    with pytest.raises(ValueError) as e:
        df_to_heatmap(df, **kwargs)
    assert str(e.value) == error



from __future__ import annotations

from datetime import datetime
from pathlib import Path

import iris
import iris.cube
import numpy as np
import pytest
import xarray as xr
from iris.cube import Cube
from numpy.testing import assert_allclose

from pyaerocom import GriddedData, Variable, const
from pyaerocom.exceptions import (
    CoordinateError,
    DataDimensionError,
    DataSearchError,
    VariableDefinitionError,
    VariableNotFoundError,
)
from pyaerocom.io import ReadGridded
from tests.conftest import TEST_RTOL, need_iris_32

TESTLATS = [-10, 20]
TESTLONS = [-120, 69]


def test_GriddedData_var_name():
    data = GriddedData()
    assert data.var_name is None
    data.var_name = "Blaaa"
    assert data.var_name == data.grid.var_name == "Blaaa"


def test_GriddedData_var_name_error():
    not_a_str = None
    with pytest.raises(ValueError) as e:
        GriddedData().var_name = not_a_str
    assert str(e.value) == f"Invalid input for var_name, need str, got {not_a_str}"


@pytest.mark.parametrize(
    "var_name, var_name_aerocom",
    [
        ("BlBlub", None),
        ("od550aer", "od550aer"),
        ("scatc550aer", "sc550aer"),
    ],
)
def test_GriddedData_var_name_aerocom(var_name, var_name_aerocom):
    data = GriddedData()
    data.var_name = var_name
    assert data.var_name_aerocom == var_name_aerocom


def test_GriddedData_var_info():
    data = GriddedData()
    data.var_name = "od550aer"
    assert isinstance(data.var_info, Variable)


def test_GriddedData_var_info_error():
    data = GriddedData()
    data.var_name = "manamana"
    with pytest.raises(VariableDefinitionError) as e:
        data.var_info
    assert str(e.value) == f"No default access available for variable {data.var_name}"


def test_GriddedData_long_name():
    data = GriddedData()
    assert data.long_name is None
    data.long_name = "blaaa"
    assert data.long_name == data.grid.long_name == "blaaa"


def test_GriddedData_suppl_info():
    assert isinstance(GriddedData().metadata, iris.cube.CubeAttrsDict)


def test_basic_properties(data_tm5: GriddedData):
    assert isinstance(data_tm5.cube, Cube)
    assert data_tm5.ts_type == "monthly"
    assert str(data_tm5.start) == "2010-01-01T00:00:00.000000"
    assert str(data_tm5.stop) == "2010-12-31T23:59:59.999999"
    assert len(data_tm5.time.points) == 12
    assert data_tm5.data_id == "TM5_AP3-CTRL2016"
    assert [Path(file).name for file in data_tm5.from_files] == [
        "aerocom3_TM5_AP3-CTRL2016_od550aer_Column_2010_monthly.nc"
    ]
    assert data_tm5.shape == (12, 90, 120)
    assert data_tm5.lat_res == 2.0
    assert data_tm5.lon_res == 3.0


@need_iris_32
def test_GriddedData_longitude(data_tm5: GriddedData):
    """Test if longitudes are defined right"""
    assert str(data_tm5.longitude.units) == "degrees"

    lons = data_tm5.longitude.points
    assert (lons.min(), lons.max()) == (-178.5, 178.5)


def test_GriddedData_latitude(data_tm5: GriddedData):
    """test latitude array"""
    assert str(data_tm5.latitude.units) == "degrees"
    lats = data_tm5.latitude.points
    assert (lats.min(), lats.max()) == (-89, 89)


def test_GriddedData_time(data_tm5: GriddedData):
    """Test time dimension access and values"""
    time = data_tm5.time

    nominal_eq = ["julian", "day since 1850-01-01 00:00:00.0000000 UTC", False]
    vals_eq = [time.units.calendar, time.units.name, isinstance(time.cell(0).point, datetime)]
    assert nominal_eq == vals_eq


def test_GriddedData_resample_time(data_tm5: GriddedData):
    yearly = data_tm5.resample_time("yearly")
    assert yearly.shape == (1, 90, 120)

    assert_allclose(data_tm5.mean(), 0.11865, rtol=TEST_RTOL)
    assert_allclose(yearly.mean(), 0.11865, rtol=TEST_RTOL)


def test_GriddedData_interpolate(data_tm5: GriddedData):
    data = data_tm5.interpolate(latitude=TESTLATS, longitude=TESTLONS)

    assert type(data) is GriddedData
    assert data.shape == (12, 2, 2)

    assert_allclose(data.mean(False), 0.13877, rtol=TEST_RTOL)
    assert_allclose(data.mean(True), 0.13748, rtol=TEST_RTOL)


def test_GriddedData_to_time_series(data_tm5: GriddedData):
    stats = data_tm5.to_time_series(latitude=TESTLATS, longitude=TESTLONS)
    assert [stat.latitude for stat in stats] == [-9, 21]
    assert [stat.longitude for stat in stats] == [-118.5, 70.5]
    assert_allclose([stat.od550aer.mean() for stat in stats], [0.101353, 0.270886], rtol=TEST_RTOL)


def test_GriddedData_change_baseyear(data_tm5: GriddedData):
    data = data_tm5.copy()
    data.change_base_year(1901)
    assert str(data.time.units) == "days since 1901-01-01 00:00:00"


def test_GriddedData_min(data_tm5: GriddedData):
    assert_allclose(data_tm5.min(), 0.004629, atol=0.0001)


def test_GriddedData_nanmin(data_tm5: GriddedData):
    assert_allclose(data_tm5.nanmin(), 0.004629, atol=0.0001)


def test_GriddedData_max(data_tm5: GriddedData):
    assert_allclose(data_tm5.max(), 2.495539, atol=0.0001)


def test_GriddedData_nanmax(data_tm5: GriddedData):
    assert_allclose(data_tm5.nanmax(), 2.495539, atol=0.0001)


@pytest.mark.parametrize(
    "extend_percent,expected",
    [
        (0, (0.004, 2.496)),
        (15, (-0.4, 2.9)),
    ],
)
def test_GriddedData_estimate_value_range_from_data(
    data_tm5: GriddedData, extend_percent: int, expected: tuple[float]
):
    result = data_tm5.estimate_value_range_from_data(extend_percent)
    assert_allclose(result, expected, rtol=1e-2)


def test_GriddedData_area_weighted_mean(data_tm5: GriddedData):
    mean = data_tm5.area_weighted_mean()
    assert len(mean) == 12
    assert_allclose(mean.mean(), 0.118648, atol=0.001)


def test_GriddedData_mean(data_tm5: GriddedData):
    assert_allclose(data_tm5.mean(areaweighted=True), 0.11864813532841474)
    assert_allclose(data_tm5.mean(areaweighted=False), 0.09825691)


def test_GriddedData_std(data_tm5: GriddedData):
    assert_allclose(data_tm5.std(), 0.106527, atol=0.0001)


def test_GriddedData_short_str(data_tm5: GriddedData):
    assert data_tm5.short_str() == "od550aer (TM5_AP3-CTRL2016, freq=monthly, unit=1)"


def test_GriddedData_copy(data_tm5: GriddedData):
    data = data_tm5.copy()
    assert isinstance(data, GriddedData)
    assert data.cube is not data_tm5.cube


def test_GriddedData__check_lonlat_bounds(data_tm5: GriddedData):
    data = data_tm5.copy()
    data.latitude.bounds = None
    data.longitude.bounds = None
    data._check_lonlat_bounds()
    lonb = data.longitude.bounds
    latb = data.latitude.bounds
    assert isinstance(latb, np.ndarray)
    assert isinstance(lonb, np.ndarray)
    assert lonb.shape == (120, 2)
    assert latb.shape == (90, 2)


@pytest.mark.parametrize(
    "val,expected",
    [
        ("lon", {"var_name": "lon"}),
        ("longitude", {"standard_name": "longitude"}),
        ("Center coordinates for longitudes", {"long_name": "Center coordinates for longitudes"}),
        ("lat", {"var_name": "lat"}),
        ("latitude", {"standard_name": "latitude"}),
        ("Center coordinates for latitudes", {"long_name": "Center coordinates for latitudes"}),
        ("time", {"standard_name": "time"}),
        ("Time", {"long_name": "Time"}),
    ],
)
def test_GriddedData__check_coordinate_access(data_tm5: GriddedData, val: str, expected: dict):
    assert data_tm5._check_coordinate_access(val) == expected


def test_GriddedData__check_coordinate_access_error(data_tm5: GriddedData):
    wrong_coord = "not_a_coordinate"
    with pytest.raises(CoordinateError) as e:
        data_tm5._check_coordinate_access(wrong_coord)
    assert (
        str(e.value)
        == f"Could not associate one of the coordinates with input string {wrong_coord}"
    )


@pytest.mark.parametrize("add_aux", [True, False])
def test_GriddedData_delete_aux_vars(data_tm5: GriddedData, add_aux: bool):
    data = data_tm5.copy()

    if add_aux:
        auxc = iris.coords.AuxCoord(data.time.points, var_name="time2")
        data.cube.add_aux_coord(auxc, [0])
        assert len(data.cube.aux_coords) == 1

    data.delete_aux_vars()
    assert len(data.cube.aux_coords) == 0


def test_GriddedData_reader_setter(data_tm5: GriddedData):
    data = data_tm5.copy()
    data.reader = reader = ReadGridded("TM5-met2010_CTRL-TEST")
    assert data._reader is reader
    assert data.reader is reader


def test_GriddedData_reader_setter_error(data_tm5: GriddedData):
    with pytest.raises(ValueError) as e:
        data_tm5.copy().reader = 24
    assert str(e.value).startswith("cannot set reader")


def test_GriddedData_reader_getter(data_tm5: GriddedData):
    data = data_tm5.copy()
    data.metadata["data_id"] = "TM5-met2010_CTRL-TEST"
    assert data._reader is None
    assert isinstance(data.reader, ReadGridded)


def test_GriddedData_reader_getter_error(data_tm5: GriddedData):
    data = data_tm5.copy()
    data.metadata["data_id"] = data_id = "blaaaa"
    with pytest.raises(DataSearchError) as e:
        data.reader
    assert str(e.value) == f"No matches could be found for search pattern {data_id}"


def test_GriddedData_search_other():
    reader = ReadGridded("TM5-met2010_CTRL-TEST")
    variable = "od550aer"
    data = reader.read_var(variable, start=2010, ts_type="monthly")
    assert isinstance(data.search_other(variable), GriddedData)


def test_GriddedData_search_other_error():
    reader = ReadGridded("TM5-met2010_CTRL-TEST")
    data = reader.read_var("od550aer", start=2010, ts_type="monthly")
    wrong_variable = "concso4"
    with pytest.raises(VariableNotFoundError) as e:
        data.search_other(wrong_variable)
    assert str(e.value) == f"Could not find variable {wrong_variable}"


def test_GriddedData_update_meta(data_tm5: GriddedData):
    data = data_tm5.copy()
    data.update_meta(bla=42, blub=43)
    assert data.metadata["bla"] == 42
    assert data.metadata["blub"] == 43


@pytest.mark.parametrize("inplace", [True, False])
def test_GriddedData_delete_all_coords(data_tm5: GriddedData, inplace: bool):
    data = data_tm5.copy()
    new = data.delete_all_coords(inplace)
    assert new.cube.coords() == []
    if inplace:
        assert data is new
    else:
        assert len(data.cube.coords()) == 3


@pytest.mark.parametrize("inplace", [True, False])
def test_GriddedData_copy_coords(inplace: bool):
    reader = ReadGridded("TM5-met2010_CTRL-TEST")
    aod = reader.read_var("od550aer", start=2010, ts_type="monthly")
    abs = reader.read_var("abs550aer", start=2010, ts_type="monthly")
    result = aod.copy_coords(abs, inplace)
    assert (result.cube is aod.cube) == inplace
    for coord in abs.cube.coords():
        assert coord == result.cube.coord(coord.name())


def test_GriddedData_copy_coords_error():
    reader = ReadGridded("TM5-met2010_CTRL-TEST")
    aod = reader.read_var("od550aer", start=2010, ts_type="monthly")
    abs = reader.read_var("abs550aer", start=2010, ts_type="daily")
    with pytest.raises(DataDimensionError) as e:
        aod.copy_coords(abs, False)
    assert str(e.value) == "Cannot copy coordinates: shape mismatch"


def test_GriddedData_register_var_glob():
    arr = np.ones((10, 10, 10))
    arr[2:5] = 4
    var_name = "blablub"
    cube = Cube(arr, var_name=var_name)
    data = GriddedData(input=cube)
    data.register_var_glob()
    vars = const.VARS
    assert vars._all_vars[-1] == var_name
    vars._all_vars.pop(-1)
    del vars._vars_added[var_name]


@pytest.fixture
def fake_dataset_path(tmp_path: Path, var_name: str, units: str) -> Path:
    arr = xr.DataArray(np.ones(10))
    arr.attrs["var_name"] = var_name
    arr.attrs["units"] = units
    ds = arr.to_dataset(name=var_name)

    path = tmp_path / "output.nc"
    ds.to_netcdf(path)
    assert path.exists()
    return path


@pytest.mark.parametrize(
    "var_name,units,data_unit",
    [
        ("od550aer", "1", "1"),
        ("concso4", "ug S m-3", "ug S m-3"),
        ("concco", "ugC/m3", "ug C m-3"),
    ],
)
# iris after 3.7 changed warnings
@pytest.mark.filterwarnings(
    "ignore::iris.fileformats._nc_load_rules.helpers._WarnComboIgnoringCfLoad"
)
def test_GriddedData__check_invalid_unit_alias(
    fake_dataset_path: Path, var_name: str, data_unit: str
):
    data = GriddedData(fake_dataset_path, var_name=var_name, check_unit=False)
    data._check_invalid_unit_alias()
    assert data.units == data_unit


def test_mean_at_coords(data_tm5: GriddedData):
    data = data_tm5.copy()
    mean = data.mean_at_coords(latitude=data.lat.points, longitude=data.lon.points)
    assert isinstance(mean, np.floating)


def test__coords_to_iris_sample_points(data_tm5: GriddedData):
    data = data_tm5.copy()
    assert len(data._coords_to_iris_sample_points()) == 0
    points = data._coords_to_iris_sample_points(coords=(data.lat, data.lon))
    assert isinstance(points, list)
    assert points[0][1][0].shape == data.lat.shape
    assert points[0][1][1].shape == data.lon.shape


def test_extract_surface_level_wrong_dim(
    data_tm5: GriddedData,
):  # can only currently test the exception gets raised b/c don't have a good 4D testing dataset
    data = data_tm5.copy()
    with pytest.raises(DataDimensionError) as e:
        data.extract_surface_level()
    assert e.type is DataDimensionError


def test__infer_index_surface_level_wrong_dim(
    data_tm5: GriddedData,
):  # can only currently test the exception gets raised b/c don't have a good 4D testing dataset
    data = data_tm5.copy()
    with pytest.raises(DataDimensionError) as e:
        data._infer_index_surface_level()
    assert e.type is DataDimensionError


def test_find_closest_index_empty(
    data_tm5: GriddedData,
):  # can't find a place where called in codebase but doesn't seem to be depricated
    empty = data_tm5.find_closest_index()
    assert len(empty) == 0


def test_remove_outliers(data_tm5: GriddedData):
    data = data_tm5.copy()
    new = data.remove_outliers(inplace=False)
    assert data.shape == new.shape
    assert new.metadata["outliers_removed"]


def test__resample_time_iris(data_tm5: GriddedData):
    data = data_tm5.copy()
    new = data._resample_time_iris("yearly")
    assert new.ts_type == "yearly"


def test_get_area_weighted_timeseries(data_tm5: GriddedData):
    data = data_tm5.copy()
    new = data.get_area_weighted_timeseries("EUROPE")
    assert new.region


def test_extract(data_tm5: GriddedData):
    data = data_tm5.copy()
    sub = data.extract(iris.Constraint(latitude=0.0))
    assert len(sub.shape) < len(data.shape)


def test_intersection(data_tm5: GriddedData):
    data = data_tm5.copy()
    new = data.intersection(longitude=(-70, 35))
    assert new.shape[0:1] == data.shape[0:1]
    assert new.shape[2] < data.shape[2]



from datetime import timedelta

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from pyaerocom import StationData, helpers
from pyaerocom.exceptions import DataCoverageError, TemporalResolutionError, UnitConversionError


def test_get_standarad_name():
    assert (
        helpers.get_standard_name("od550aer")
        == "atmosphere_optical_thickness_due_to_ambient_aerosol_particles"
    )


def test_get_standard_unit():
    assert helpers.get_standard_unit("ec550aer") == "1/km"


def test_get_lowest_resolution():
    assert helpers.get_lowest_resolution("3hourly", "hourly", "monthly", "yearly") == "yearly"


@pytest.mark.parametrize("val", [3, 3.3455, complex(1, 2)])
def test_isnumeric(val):
    assert helpers.isnumeric(val)


@pytest.mark.parametrize(
    "val,result",
    [
        ((0, 1), True),
        ([10, 20], True),
        ([10, 20, 30], False),
    ],
)
def test_isrange(val, result):
    assert helpers.isrange(val) == result


@pytest.mark.parametrize(
    "use,var_name,pref_attr,sort_by_largest,fill_missing_nan,add_meta_keys,num,tst,mean",
    [
        ("concpm10", "concpm10", None, True, True, None, 730, "daily", 17.93),
        ("concpm10", "concpm10", "awesomeness", True, True, None, 730, "daily", 17.93),
        ("concpm10", "concpm10", "awesomeness", False, True, None, 730, "daily", 15),
        ("od550aer", "od550aer", None, True, True, None, 67, "60daily", 0.51),
        ("od550aer", "od550aer", "awesomeness", True, True, None, 67, "60daily", 0.59),
        ("od550aer", "od550aer", "awesomeness", False, True, None, 67, "60daily", 0.51),
    ],
)
def test_merge_station_data(
    statlist,
    use,
    var_name,
    pref_attr,
    sort_by_largest,
    fill_missing_nan,
    add_meta_keys,
    num,
    tst,
    mean,
):
    stats = [x.copy() for x in statlist[use]]
    stat = helpers.merge_station_data(
        stats, var_name, pref_attr, sort_by_largest, fill_missing_nan, add_meta_keys
    )
    assert isinstance(stat, StationData)
    vardata = stat[var_name]
    assert len(vardata) == num
    assert stat.get_var_ts_type(var_name) == tst
    assert np.mean(vardata) == pytest.approx(mean, rel=1e-2)


@pytest.mark.parametrize(
    "use,exception,error",
    [
        ("concpm10_X2", UnitConversionError, "failed to convert unit from mole mole-1 to ug m-3"),
        ("concpm10_X", TemporalResolutionError, "Invalid input for ts_type daily"),
        ("od550aer", DataCoverageError, "All input stations must contain concpm10 data"),
    ],
)
def test_merge_station_data_error(statlist, use, exception, error):
    stats = [x.copy() for x in statlist[use]]
    with pytest.raises(exception) as e:
        helpers.merge_station_data(stats, "concpm10")
    assert str(e.value).startswith(error)


def test__get_pandas_freq_and_offset():
    val = helpers._get_pandas_freq_and_offset("monthly")
    assert val == ("MS", pd.Timedelta(14, "d"))


@pytest.fixture(scope="module")
def fake_hourly_ts():
    time = pd.date_range("2018-01-10T00:00:00", "2018-01-17T23:59:00", freq="h")
    xrange_modulation = np.linspace(0, np.pi * 40, len(time))
    signal = np.sin(xrange_modulation)
    return pd.Series(signal, time)


@pytest.mark.parametrize(
    "freq,how,min_num_obs,num,avg",
    [
        ("daily", "mean", 10000, 8, np.nan),
        ("daily", "50percentile", 10000, 8, np.nan),
        ("daily", "mean", 23, 8, 0),
        ("daily", "50percentile", 23, 8, 0),
        ("yearly", "50percentile", 23, 1, 0),
        ("yearly", "50percentile", 192, 1, 0),
        ("daily", "50percentile", None, 8, 0),
        ("yearly", "50percentile", 193, 1, np.nan),
        ("daily", "median", None, 8, 0.0),
        ("monthly", "mean", None, 1, 0),
        ("daily", "mean", None, 8, 0),
        ("daily", "1percentile", None, 8, -1),
        ("daily", "25percentile", None, 8, -0.64),
        ("daily", "75percentile", None, 8, 0.64),
    ],
)
@pytest.mark.filterwarnings("ignore:Mean of empty slice:RuntimeWarning")
def test_resample_timeseries(fake_hourly_ts, freq, how, min_num_obs, num, avg):
    s1 = helpers.resample_timeseries(fake_hourly_ts, freq=freq, how=how, min_num_obs=min_num_obs)
    assert len(s1) == num
    assert np.nanmean(s1) == pytest.approx(avg, abs=1e-2, nan_ok=True)


def test_same_meta_dict():
    d1 = dict(
        station_name="bla", station_id="blub", latitude=33, longitude=15, altitude=400, PI="pi1"
    )
    d2 = dict(
        station_name="bla", station_id="blub1", latitude=33, longitude=15, altitude=401, PI="pi2"
    )

    assert helpers.same_meta_dict(d1, d2) is False


def test_to_pandas_timestamp():
    pass


def test_to_datetime64():
    pass


def test_is_year():
    assert helpers.is_year(2010)


def test_start_stop():
    pass


def test_datetime2str():
    pass


def test_start_stop_str():
    pass


def test_start_stop_from_year():
    start, stop = helpers.start_stop_from_year(2000)
    assert start == pd.Timestamp("2000")
    assert stop == pd.Timestamp("2000-12-31 23:59:59")


@pytest.mark.parametrize(
    "input,expected",
    [("20100101", "20100101"), (helpers.to_pandas_timestamp("20100101"), "20100101")],
)
def test_to_datestring_YYYYMMDD(input, expected):
    assert helpers.to_datestring_YYYYMMDD(input) == expected


def test_cftime_to_datetime64():
    pass


def get_constraint():
    pass


def test_get_lat_rng_constraint():
    pass


def test_get_lon_rng_constraint():
    pass


def test_get_time_rng_constraint():
    pass


def test_extract_latlon_dataarray():
    cube = helpers.make_dummy_cube_latlon(
        lat_res_deg=1, lon_res_deg=1, lat_range=[10, 20], lon_range=[10, 20]
    )
    data = xr.DataArray.from_iris(cube)
    # First coordinate does not exist in the dataarray.
    lat = [15, 15, 18]
    lon = [1, 15, 18]
    subset = helpers.extract_latlon_dataarray(data, lat, lon, check_domain=True)
    assert isinstance(subset, xr.DataArray)
    assert len(subset.lat) == len(lat) - 1
    assert len(subset.lon) == len(lon) - 1


def test_extract_latlon_dataarray_no_matches():
    cube = helpers.make_dummy_cube_latlon(
        lat_res_deg=1, lon_res_deg=1, lat_range=[10, 20], lon_range=[10, 20]
    )
    data = xr.DataArray.from_iris(cube)
    helpers.extract_latlon_dataarray(data, [15], [15], check_domain=True)


@pytest.mark.parametrize(
    "lat,lon",
    [
        ([], []),
        ([1, 2], [-1, 2]),
    ],
)
def test_extract_latlon_dataarray_no_matches_error(lat, lon):
    cube = helpers.make_dummy_cube_latlon(
        lat_res_deg=1, lon_res_deg=1, lat_range=[10, 20], lon_range=[10, 20]
    )
    data = xr.DataArray.from_iris(cube)
    with pytest.raises(DataCoverageError) as e:
        helpers.extract_latlon_dataarray(data, lat, lon, check_domain=True)
    assert str(e.value) == "Coordinates not found in dataarray"


@pytest.mark.parametrize(
    "date,ts_type,days",
    [
        pytest.param("2000-02-18", "yearly", 366, id="leap year"),
        pytest.param("2001-02-18", "yearly", 365, id="lon leap year"),
        pytest.param("2000-02-18", "monthly", 29, id="February leap year"),
        pytest.param("2001-02-18", "monthly", 28, id="February non leap year"),
        pytest.param("2001-02-18", "daily", 1, id="one day"),
    ],
)
def test_seconds_in_periods(date, ts_type, days):
    seconds = timedelta(days=days) / timedelta(seconds=1)
    assert helpers.seconds_in_periods(np.datetime64(date), ts_type) == seconds


def test_make_dummy_cube():
    # make a dummy cube of an arbitrary variable name over one year
    cube = helpers.make_dummy_cube("concpm10", start_yr=2020, stop_yr=2021)
    data = xr.DataArray.from_iris(cube)
    # First coordinate does not exist in the dataarray.
    lat = [90, -67.5, 22.2]
    lon = [180, -135.0, 45.0]
    subset = helpers.extract_latlon_dataarray(data, lat, lon, check_domain=True)
    assert isinstance(subset, xr.DataArray)
    assert len(subset.lat) == len(lat) - 1
    assert len(subset.lon) == len(lon) - 1



from pathlib import Path

import numpy as np
import pytest
from pydantic import ValidationError

from pyaerocom import ColocatedData, GriddedData, UngriddedData, const
from pyaerocom.colocation.colocation_setup import ColocationSetup
from pyaerocom.colocation.colocator import Colocator
from pyaerocom.config import ALL_REGION_NAME
from pyaerocom.exceptions import ColocationError, ColocationSetupError
from pyaerocom.io.aux_read_cubes import add_cubes
from pyaerocom.io.mscw_ctm.reader import ReadMscwCtm
from tests.fixtures.data_access import TEST_DATA

COL_OUT_DEFAULT = Path(const.OUTPUTDIR) / "colocated_data"

default_setup = {
    "model_id": None,
    "obs_id": None,
    "obs_vars": (),
    "ts_type": "monthly",
    "start": None,
    "stop": None,
    "filter_name": f"{ALL_REGION_NAME}-wMOUNTAINS",
    "basedir_coldata": COL_OUT_DEFAULT,
    "save_coldata": False,
    "obs_name": None,
    "obs_data_dir": None,
    "obs_use_climatology": False,
    "obs_cache_only": False,
    "obs_vert_type": None,
    "obs_ts_type_read": None,
    "obs_filters": {},
    "model_name": None,
    "model_data_dir": None,
    "model_read_opts": {},
    "read_opts_ungridded": {},
    "model_use_vars": {},
    "model_rename_vars": {},
    "model_add_vars": {},
    "model_to_stp": False,
    "model_ts_type_read": None,
    "model_read_aux": {},
    "model_use_climatology": False,
    "gridded_reader_id": {"model": "ReadGridded", "obs": "ReadGridded"},
    "flex_ts_type": True,
    "min_num_obs": None,
    "resample_how": "mean",
    "obs_remove_outliers": False,
    "model_remove_outliers": False,
    "zeros_to_nan": False,
    "obs_outlier_ranges": {},
    "model_outlier_ranges": {},
    "harmonise_units": False,
    "regrid_res_deg": None,
    "colocate_time": False,
    "reanalyse_existing": True,
    "raise_exceptions": False,
    "keep_data": True,
    "add_meta": {},
}


@pytest.fixture(scope="function")
def setup():
    return dict(
        model_id="TM5-met2010_CTRL-TEST",
        obs_id="AeronetSunV3L2Subset.daily",
        obs_vars="od550aer",
        start=2010,
        raise_exceptions=True,
        reanalyse_existing=True,
    )


@pytest.fixture(scope="function")
def tm5_aero_col_stp(setup):
    return ColocationSetup(**setup)


@pytest.fixture(scope="function")
def col(tm5_aero_col_stp):
    col = Colocator(tm5_aero_col_stp, raise_exceptions=True, reanalyse_existing=True)
    return col


@pytest.mark.parametrize("col_stp,should_be", [(ColocationSetup(), default_setup)])
def test_colocation_setup(col_stp: ColocationSetup, should_be: dict):
    stp_dict = col_stp.model_dump()
    for key, val in should_be.items():
        assert key in stp_dict
        if key == "basedir_coldata":
            assert Path(val) == Path(stp_dict["basedir_coldata"])
        else:
            assert val == stp_dict[key], key


@pytest.mark.parametrize(
    "key,val,raises",
    [
        ("obs_vars", 42, pytest.raises(ValidationError)),
        ("var_ref_outlier_ranges", [41, 42], pytest.raises(KeyError)),
        ("var_outlier_ranges", [41, 42], pytest.raises(KeyError)),
        ("remove_outliers", True, pytest.raises(KeyError)),
    ],
)
def test_ColocationSetup_invalid_input(key, val, raises):
    with raises:
        stp = ColocationSetup(**{key: val})
        assert stp.model_dump()[key] == val


@pytest.mark.parametrize("ts_type_desired", ["daily", "monthly"])
@pytest.mark.parametrize("ts_type", ["monthly"])
@pytest.mark.parametrize("flex", [False, True])
def test_Colocator_model_ts_type_read(setup, ts_type_desired, ts_type, flex):
    obs_var = "od550aer"
    setup["model_ts_type_read"] = {obs_var: ts_type_desired}
    setup.update(dict(ts_type=ts_type, flex=flex))
    assert setup["obs_vars"] == obs_var
    col_stp = ColocationSetup(**setup)
    col = Colocator(col_stp)
    # Problem with saving since obs_id is different
    # from obs_data.contains_dataset[0]...
    data = col.run()
    assert isinstance(data, dict)
    assert obs_var in data
    coldata = data[obs_var][obs_var]
    assert coldata.ts_type == ts_type
    assert coldata.metadata["ts_type_src"][0] == "daily"
    if not flex:
        assert coldata.metadata["ts_type_src"][1] == ts_type_desired


def test_Colocator_model_ts_type_read_error(setup):
    setup["model_ts_type_read"] = {"od550aer": "minutely"}
    col_stp = ColocationSetup(**setup)
    col = Colocator(col_stp)
    with pytest.raises(ColocationError) as e:
        col.run()
    assert str(e.value).startswith("Failed to load model data: TM5-met2010_CTRL-TEST (od550aer)")


def test_Colocator_model_add_vars(setup):
    model_var = "abs550aer"
    obs_var = "od550aer"
    setup["model_add_vars"] = {obs_var: [model_var]}
    col_stp = ColocationSetup(**setup)
    col = Colocator(col_stp)
    # Problem with saving since obs_id is different
    data = col.run(var_list=[model_var])
    assert isinstance(data, dict)
    assert model_var in data
    coldata = data[model_var][obs_var]
    assert coldata.var_name == ["od550aer", "abs550aer"]


def test_Colocator_init_basedir_coldata(setup, tmp_path: Path):
    base_path = tmp_path / "basedir"
    setup["basedir_coldata"] = base_path
    setup["raise_exceptions"] = True
    col_stp = ColocationSetup(**setup)
    Colocator(col_stp)

    assert base_path.is_dir()


def test_Colocator__infer_start_stop_yr_from_model_reader(tm5_aero_col_stp):
    col = Colocator(tm5_aero_col_stp)
    col.model_id = "TM5-met2010_CTRL-TEST"
    col._infer_start_stop_yr_from_model_reader()
    assert col.start == 2010
    assert col.stop is None


def test_Colocator__coldata_savename(setup):
    setup["raise_exceptions"] = True
    setup["obs_name"] = "obs"
    setup["model_name"] = "model"
    setup["filter_name"] = ALL_REGION_NAME
    setup["start"] = 2015
    col_stp = ColocationSetup(**setup)
    col = Colocator(col_stp)
    col._check_set_start_stop()
    savename = col._coldata_savename("od550aer", "od550ss", "daily")
    assert isinstance(savename, str)
    n = f"od550ss_od550aer_MOD-model_REF-obs_20150101_20151231_daily_{ALL_REGION_NAME}.nc"
    assert savename == n


def test_Colocator_run_gridded_gridded(setup):
    setup["obs_id"] = setup["model_id"]
    col_stp = ColocationSetup(**setup)
    col = Colocator(col_stp)
    col.run()
    var = col.colocation_setup.obs_vars[0]
    coldata = col.data[var][var]
    assert isinstance(coldata, ColocatedData)
    assert coldata.ndim == 4


@pytest.mark.parametrize(
    "update,chk_mvar,chk_ovar,sh,mean_obs,mean_mod",
    [
        (dict(), "od550aer", "od550aer", (2, 12, 11), 0.272, 0.244),
        (dict(regrid_res_deg=10), "od550aer", "od550aer", (2, 12, 11), 0.272, 0.229),
        (dict(), "od550aer", "od550aer", (2, 12, 11), 0.272, 0.244),
        (
            dict(
                model_use_vars={"od550aer": "abs550aer"},
                model_use_climatology=True,
                obs_use_climatology=True,
            ),
            "abs550aer",
            "od550aer",
            (2, 12, 1),
            0.123,
            0.002,
        ),
        (
            dict(model_use_vars={"od550aer": "abs550aer"}, model_use_climatology=True),
            "abs550aer",
            "od550aer",
            (2, 12, 1),
            0.159,
            0.002,
        ),
        (
            dict(model_use_vars={"od550aer": "abs550aer"}, obs_use_climatology=True),
            "abs550aer",
            "od550aer",
            (2, 12, 16),
            0.259,
            0.014,
        ),
    ],
)
def test_Colocator_run_gridded_ungridded(
    setup, update, chk_mvar, chk_ovar, sh, mean_obs, mean_mod
):
    setup.update(update)
    col_stp = ColocationSetup(**setup)

    result = Colocator(col_stp).run()
    assert isinstance(result, dict)

    coldata = result[chk_mvar][chk_ovar]
    assert coldata.shape == sh

    mod_clim_used = any("9999" in x for x in coldata.metadata["from_files"])
    assert col_stp.model_use_climatology == mod_clim_used

    assert np.nanmean(coldata.data[0].values) == pytest.approx(mean_obs, abs=0.01)
    assert np.nanmean(coldata.data[1].values) == pytest.approx(mean_mod, abs=0.01)


@pytest.mark.parametrize(
    "update,error",
    [
        pytest.param(dict(obs_vars=[]), "no observation variables specified", id="no obs"),
        pytest.param(
            dict(
                model_use_vars={"od550aer": "abs550aer"},
                model_use_climatology=True,
                obs_use_climatology=True,
                start=2008,
                stop=2012,
            ),
            "Conflict: only single year analyses are support",
            id="unsupported",
        ),
    ],
)
def test_Colocator_run_gridded_ungridded_error(setup, update, error):
    setup.update(update)
    col_stp = ColocationSetup(**setup)
    with pytest.raises(ColocationSetupError) as e:
        Colocator(col_stp).run()
    assert str(e.value).startswith(error)


def test_colocator_filter_name(setup):
    setup["filter_name"] = ALL_REGION_NAME
    col_stp = ColocationSetup(**setup)
    col = Colocator(col_stp)
    assert col.colocation_setup.filter_name == ALL_REGION_NAME


def test_colocator_read_ungridded(setup):
    # obs_id = "AeronetSunV3L2Subset.daily"
    obs_var = "od550aer"
    # setup["obs_id"] =
    setup["raise_exceptions"] = True
    setup["obs_filters"] = {"longitude": [-30, 30]}
    setup["read_opts_ungridded"] = {"last_file": 1}
    col_stp = ColocationSetup(**setup)
    col = Colocator(col_stp)

    data = col._read_ungridded(obs_var)
    assert isinstance(data, UngriddedData)
    assert len(data.metadata) == 1

    col.obs_vars = ["invalid"]
    with pytest.raises(ValueError):
        data = col._read_ungridded("invalid")


def test_colocator_get_model_data(setup):
    setup["raise_exceptions"] = True
    setup["model_id"] = "TM5-met2010_CTRL-TEST"
    col_stp = ColocationSetup(**setup)

    col = Colocator(col_stp)
    data = col.get_model_data("od550aer")
    assert isinstance(data, GriddedData)


def test_colocator__find_var_matches(setup):
    setup["model_id"] = "TM5-met2010_CTRL-TEST"
    setup["obs_id"] = "AeronetSunV3L2Subset.daily"
    setup2 = setup.copy()
    setup["obs_vars"] = "od550aer"
    col_stp = ColocationSetup(**setup)
    col = Colocator(col_stp)

    var_matches = col._find_var_matches()
    assert var_matches == {"od550aer": "od550aer"}

    obs_var = "conco3"
    setup2["obs_vars"] = obs_var
    setup2["model_use_vars"] = {obs_var: "od550aer"}
    col_stp2 = ColocationSetup(**setup2)
    col2 = Colocator(col_stp2)

    var_matches = col2._find_var_matches()
    assert var_matches == {"od550aer": "conco3"}


def test_colocator__find_var_matches_model_add_vars(setup):
    ovar = "od550aer"
    setup["model_id"] = "TM5-met2010_CTRL-TEST"
    setup["obs_id"] = "AeronetSunV3L2Subset.daily"
    setup["obs_vars"] = (ovar,)
    setup["model_add_vars"] = {ovar: ("abs550aer",)}
    col_stp = ColocationSetup(**setup)
    col = Colocator(col_stp)
    var_matches = col._find_var_matches()
    assert var_matches == {"abs550aer": ovar, ovar: ovar}


def test_colocator_instantiate_gridded_reader(setup, path_emep):
    model_id = "model"
    setup["gridded_reader_id"] = {"model": "ReadMscwCtm", "obs": "ReadGridded"}
    setup["model_id"] = model_id
    setup["filepath"] = path_emep
    col_stp = ColocationSetup(**setup)
    col = Colocator(col_stp)
    r = col._instantiate_gridded_reader(what="model")
    assert isinstance(r, ReadMscwCtm)
    assert r.data_id == model_id


def test_colocator_instantiate_gridded_reader_model_data_dir(setup, path_emep):
    model_data_dir = path_emep["data_dir"]
    model_id = "model"
    setup["gridded_reader_id"] = {"model": "ReadMscwCtm", "obs": "ReadGridded"}
    setup["model_data_dir"] = model_data_dir
    setup["model_id"] = model_id
    col_stp = ColocationSetup(**setup)
    col = Colocator(col_stp)
    r = col._instantiate_gridded_reader(what="model")
    assert isinstance(r, ReadMscwCtm)
    assert r._data_dir == model_data_dir
    assert r.data_id == model_id


def test_colocator__get_gridded_reader_class(setup):
    setup["gridded_reader_id"] = {"model": "ReadMscwCtm", "obs": "ReadMscwCtm"}
    col_stp = ColocationSetup(**setup)
    col = Colocator(col_stp)
    for what in ["model", "obs"]:
        assert col._get_gridded_reader_class(what=what) == ReadMscwCtm


def test_colocator__check_add_model_read_aux(setup):
    setup["raise_exceptions"] = True
    setup["model_id"] = "TM5-met2010_CTRL-TEST"
    col_stp = ColocationSetup(**setup)
    col = Colocator(col_stp)
    assert not col._check_add_model_read_aux("od550aer")
    setup["model_read_aux"] = {
        "od550aer": dict(vars_required=["od550aer", "od550aer"], fun=add_cubes)
    }
    col_stp2 = ColocationSetup(**setup)
    col2 = Colocator(col_stp2)
    assert col2._check_add_model_read_aux("od550aer")


def test_colocator_with_obs_data_dir_ungridded(setup):
    setup["obs_id"] = "AeronetSunV3L2Subset.daily"
    setup["model_id"] = "TM5-met2010_CTRL-TEST"
    setup["obs_vars"] = "od550aer"
    setup["obs_data_dir"] = TEST_DATA["AeronetSunV3L2Subset.daily"].path
    col_stp = ColocationSetup(**setup)
    col = Colocator(col_stp)
    data = col.run()
    assert len(data) == 1
    cd = data["od550aer"]["od550aer"]
    assert isinstance(cd, ColocatedData)
    assert cd.ts_type == "monthly"
    assert str(cd.start) == "2010-01-15T00:00:00.000000000"
    assert str(cd.stop) == "2010-12-15T00:00:00.000000000"


def test_colocator_with_model_data_dir_ungridded(setup):
    setup["model_id"] = "TM5-met2010_CTRL-TEST"
    setup["obs_id"] = "AeronetSunV3L2Subset.daily"
    setup["obs_vars"] = "od550aer"
    setup["model_data_dir"] = TEST_DATA["MODELS"].path / "TM5-met2010_CTRL-TEST/renamed"
    col_stp = ColocationSetup(**setup)
    col = Colocator(col_stp)
    data = col.run()
    assert len(data) == 1
    cd = data["od550aer"]["od550aer"]
    assert isinstance(cd, ColocatedData)
    assert cd.ts_type == "monthly"
    assert str(cd.start) == "2010-01-15T00:00:00.000000000"
    assert str(cd.stop) == "2010-12-15T00:00:00.000000000"


def test_colocator_with_obs_data_dir_gridded(setup):
    setup["model_id"] = "TM5-met2010_CTRL-TEST"
    setup["obs_id"] = "TM5-met2010_CTRL-TEST"
    setup["obs_vars"] = "od550aer"
    setup["obs_dir"] = TEST_DATA["MODELS"].path / "TM5-met2010_CTRL-TEST/renamed"
    col_stp = ColocationSetup(**setup)
    col = Colocator(col_stp)
    data = col.run()
    assert len(data) == 1
    cd = data["od550aer"]["od550aer"]
    assert isinstance(cd, ColocatedData)
    assert cd.ts_type == "monthly"
    assert str(cd.start) == "2010-01-15T12:00:00.000000000"
    assert str(cd.stop) == "2010-12-15T12:00:00.000000000"


###################################
#   Test for colocation with Pyaro
###################################


def test_colocation_pyaro(pyaro_testconfig, fake_aod_MSCWCtm_data_monthly_2010, setup) -> None:
    config = pyaro_testconfig[0]
    setup["pyaro_config"] = config
    setup["model_id"] = "EMEP"
    setup["gridded_reader_id"] = {"model": "ReadMscwCtm"}
    setup["model_data_dir"] = fake_aod_MSCWCtm_data_monthly_2010
    setup["obs_vars"] = "od550aer"  # This obs does not exist in Aeronet

    col_stp = ColocationSetup(**setup)
    col = Colocator(col_stp)

    data = col.run()

    cd = data["od550aer"]["od550aer"]
    assert isinstance(cd, ColocatedData)
    assert cd.ts_type == "monthly"
    assert str(cd.start) == "2010-01-15T00:00:00.000000000"
    assert str(cd.stop) == "2010-12-15T00:00:00.000000000"

    assert np.sum(np.isnan(cd.data[0, :].data)) == 0

    assert cd.data[0, :].data.shape[0] == 12



"""
Module for geographical calculations

This module contains low-level methods to perform geographical calculations,
(e.g. distance between two coordinates)
"""

import logging
import os
from copy import deepcopy

import geonum
import numpy as np
from geocoder_reverse_natural_earth import Geocoder_Reverse_Exception, Geocoder_Reverse_NE

from pyaerocom import const
from pyaerocom.helpers import isnumeric

logger = logging.getLogger(__name__)


def calc_latlon_dists(latref, lonref, latlons):
    """
    Calculate distances of (lat, lon) coords to input lat, lon coordinate

    Parameters
    ----------
    latref : float
        latitude of reference coordinate
    lonref : float
        longitude of reference coordinate
    latlons : list
        list of (lat, lon) tuples for which distances to (latref, lonref) are
        computed

    Returns
    -------
    list
        list of computed geographic distances to input reference coordinate
        for all (lat, lon) coords in `latlons`

    """
    return [haversine(latref, lonref, c[0], c[1]) for c in latlons]


def find_coord_indices_within_distance(latref, lonref, latlons, radius=1):
    """
    Find indices of coordinates that match input coordinate

    Parameters
    ----------
    latref : float
        latitude of reference coordinate
    lonref : float
        longitude of reference coordinate
    latlons : list
        list of (lat, lon) tuples for which distances to (latref, lonref) are
        computed
    radius : float or int, optional
        Maximum allowed distance to input coordinate. The default is 1.

    Returns
    -------
    ndarray
        Indices of latlon coordinates in :param:`latlons` that are within
        the specified radius around (`latref`, `lonref`). The indices are
        sorted by distance to the input coordinate, starting with the
        closest

    """
    dists = np.asarray(calc_latlon_dists(latref, lonref, latlons))
    within_tol = np.where(dists < radius)[0]
    # the following statement sorts all indices in dists that are within
    # the tolerance radius, so the first entry in the returned aaray is the
    # index of the closest coordinate within the radius and the last is the
    # furthest
    return within_tol[np.argsort(dists[within_tol])]


def get_country_info_coords(coords):
    """
    Get country information for input lat/lon coordinates

    Parameters
    ----------
    coords : list or tuple
        list of coord tuples (lat, lon) or single coord tuple

    Raises
    ------
    ValueError
        if input format is incorrect

    Returns
    -------
    list
        list of dictionaries containing country information for each input
        coordinate
    """
    if isinstance(coords, np.ndarray):
        coords = list(coords)
    if not isinstance(coords, list | tuple):
        raise ValueError("Invalid input for coords, need list or tuple or array")

    geo = Geocoder_Reverse_NE()
    ret_list = []
    # that's what reverse_geocoder used to return
    # (more a list of this)
    ret_proto = {"city": "", "country_code": "", "code": ""}

    if len(coords) == 0:  # no coordinates
        pass
    elif isnumeric(coords[0]) and len(coords) == 2:  # only one coordinate
        lat, lon = coords
        try:
            dummy = geo.lookup(lat, lon)
        except Geocoder_Reverse_Exception:
            dummy = geo.lookup_nearest(lat, lon)
            if dummy is None:
                return [ret_proto]
        # return [rg.get(coords)]
        ret_dummy = deepcopy(ret_proto)
        ret_dummy["country"] = dummy["NAME"]
        ret_dummy["country_code"] = dummy["ISO_A2_EH"]
        return [ret_dummy]
    else:
        for coord in coords:
            ret_dummy = deepcopy(ret_proto)
            lat, lon = coord
            try:
                dummy = geo.lookup(lat, lon)
            except Geocoder_Reverse_Exception:
                dummy = geo.lookup_nearest(lat, lon)
            if dummy is not None:
                ret_dummy["country"] = dummy["NAME"]
                ret_dummy["country_code"] = dummy["ISO_A2_EH"]

            ret_list.append(ret_dummy)
    return ret_list
    # return rg.search(coords)


def get_topo_data(
    lat0, lon0, lat1=None, lon1=None, topo_dataset="srtm", topodata_loc=None, try_etopo1=False
):
    """Retrieve topographic altitude for a certain location

    Supports topography datasets supported by geonum.
    These are currently (20 Feb. 19) srtm (SRTM dataset, default, automatic access if online) and
    etopo1 (ETOPO1 dataset, lower resolution, must be available on local machine or server).

    Parameters
    ----------
    lat0 : float
            start longitude for data extraction
    lon0 : float
        start latitude for data extraction
    lat1 : float
        stop longitude for data extraction (default: None). If None only
        data around lon0, lat0 will be extracted.
    lon1 : float
        stop latitude for data extraction (default: None).
        If None only data around lon0, lat0 will be extracted
    topo_dataset : str
        name of topography dataset
    topodata_loc : str
        filepath or directory containing supported topographic datasets
    try_etopo1 : bool
        if True and if access fails via input arg `topo_dataset`, then try
        to access altitude using ETOPO1 dataset.

    Returns
    -------
    geonum.TopoData
        data object containing topography data in specified range

    Raises
    ------
    ValueError
        if altitude data cannot be accessed
    """
    if topodata_loc is None:
        if topo_dataset in const.SUPPLDIRS and os.path.exists(const.SUPPLDIRS[topo_dataset]):
            topodata_loc = const.SUPPLDIRS[topo_dataset]
            logger.info(f"Found default location for {topo_dataset} topodata at\n{topodata_loc}")

    try:
        access = geonum.TopoDataAccess(topo_dataset, local_path=topodata_loc)
        topodata = access.get_data(lat0, lon0, lat1, lon1)

        return topodata
    except Exception as e:
        if try_etopo1 and not topo_dataset == "etopo1":
            logger.warning(
                f"Failed to access topography data for {topo_dataset}. "
                f"Trying ETOPO1.\nError: {repr(e)}"
            )
            return get_topo_data(
                lat0,
                lon0,
                lat1,
                lon1,
                topo_dataset="etopo1",
                topodata_loc=topodata_loc,
                try_etopo1=False,
            )
        raise


def get_topo_altitude(lat, lon, topo_dataset="srtm", topodata_loc=None, try_etopo1=True):
    """Retrieve topographic altitude for a certain location

    Supports topography datasets supported by geonum.
    These are currently (20 Feb. 19) srtm (SRTM dataset, default, automatic access if online) and
    etopo1 (ETOPO1 dataset, lower resolution, must be available on local machine or server).

    Parameters
    ----------
    lat : float
        latitude of coordinate
    lon : float
        longitude of coordinate
    topo_dataset : str
        name of topography dataset
    topodata_loc : str
        filepath or directory containing supported topographic datasets
    try_etopo1 : bool
        if True and if access fails via input arg `topo_dataset`, then try
        to access altitude using ETOPO1 dataset.

    Returns
    -------
    dict
        dictionary containing input latitude, longitude, altitude and
        topographic dataset name used to retrieve the altitude.

    Raises
    ------
    ValueError
        if altitude data cannot be accessed
    """
    return get_topo_data(
        lat, lon, topo_dataset=topo_dataset, topodata_loc=topodata_loc, try_etopo1=try_etopo1
    )(lat, lon)


def calc_distance(lat0, lon0, lat1, lon1, alt0=None, alt1=None, auto_altitude_srtm=False):
    """Calculate distance between two coordinates

    Parameters
    ----------
    lat0 : float
        latitude of first point in decimal degrees
    lon0 : float
        longitude of first point in decimal degrees
    lat1 : float
        latitude of secondpoint in decimal degrees
    lon1 : float
        longitude of second point in decimal degrees
    alt0 : :obj:`float`, optional
        altitude of first point in m
    alt1 : :obj:`float`, optional
        altitude of second point in m
    auto_altitude_srtm : bool
        if True, then all altitudes that are unspecified are set to the
        corresponding topographic altitude of that coordinate, using SRTM
        (only works for coordinates where SRTM topographic data is accessible).

    Returns
    -------
    float
        distance between points in km
    """
    p0 = geonum.GeoPoint(lat0, lon0, alt0, auto_topo_access=auto_altitude_srtm)
    p1 = geonum.GeoPoint(lat1, lon1, alt1, auto_topo_access=auto_altitude_srtm)
    if auto_altitude_srtm and p0.altitude_err == p0._ALTERR_DEFAULT:
        raise ValueError(
            f"Failed to access topographic height for coord {p0} using SRTM topographic database"
        )
    if auto_altitude_srtm and p1.altitude_err == p1._ALTERR_DEFAULT:
        raise ValueError(
            f"Failed to access topographic height for coord {p1} using SRTM topographic database"
        )
    return (p0 - p1).magnitude


def is_within_radius_km(lat0, lon0, lat1, lon1, maxdist_km, alt0=0, alt1=0, **kwargs):
    """Checks if two lon/lat coordinates are within a certain distance to each other

    Parameters
    ----------
    lat0 : float
        latitude of first point in decimal degrees
    lon0 : float
        longitude of first point in decimal degrees
    lat1 : float
        latitude of second point in decimal degrees
    lon1 : float
        longitude of second point in decimal degrees
    maxdist_km : float
        maximum distance between two points in km
    alt0 : float
        altitude of first point in m
    alt1 : float
        altitude of second point in m

    Returns
    -------
    bool
        True, if coordinates are within specified distance to each other, else
        False

    """
    dist = calc_distance(lat0, lon0, lat1, lon1, alt0=alt0, alt1=alt1)
    if dist <= maxdist_km:
        return True
    return False


def haversine(lat0, lon0, lat1, lon1, earth_radius=6371.0):
    """Haversine formula

    Approximate horizontal distance between 2 points assuming a spherical
    earth using haversine formula.

    Note
    ----
    This code was copied from geonum library (date 12/11/2018, J. Gliss)

    Parameters
    ----------
    lat0 : float
        latitude of first point in decimal degrees
    lon0 : float
        longitude of first point in decimal degrees
    lat1 : float
        latitude of second point in decimal degrees
    lon1 : float
        longitude of second point in decimal degrees
    earth_radius : float
        average earth radius in km, defaults to 6371.0

    Returns
    --------
    float
        horizontal distance between input coordinates in km
    """
    hav = lambda d_theta: np.sin(d_theta / 2.0) ** 2  # noqa: E731

    d_lon = np.radians(lon1 - lon0)
    d_lat = np.radians(lat1 - lat0)
    lat0 = np.radians(lat0)
    lat1 = np.radians(lat1)

    a = hav(d_lat) + np.cos(lat0) * np.cos(lat1) * hav(d_lon)
    c = 2 * np.arcsin(np.sqrt(a))

    return earth_radius * c



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import FuncFormatter
from seaborn import heatmap

from pyaerocom.mathutils import exponent
import warnings


def _format_annot_heatmap(annot, annot_fmt_rows, annot_fmt_exceed):
    """
    Process annotation formatting info for :func:`df_to_heatmap`

    Parameters
    ----------
    annot : ndarray
        2D pre-existing annotation information.
    annot_fmt_rows : list
        annotation formatting strings for each row of the input table. This
        parameter is only considered if `annot_fmt_rowwise` is True. See also
        :func:`_format_annot_heatmap`.
    annot_fmt_exceed : list, optional
        how to format annotated values that exceed a certain threshold.
        The list contains 2 entries, 1. the threshold values, 2. how values
        exceeding this threshold should be formatted. .

    Returns
    -------
    ndarray
        updated annotation information (input `annot`)
    list
        row formatting info.


    """
    _annot = []
    if not isinstance(annot_fmt_rows, list):
        annot_fmt_rows = []
        for row in annot:
            mask = row[~np.isnan(row)]
            if len(mask) == 0:  # all NaN
                annot_fmt_rows.append("")
                continue
            mask = mask[mask != 0]
            exps = exponent(mask)
            minexp = exps.min()
            if minexp < -3:
                annot_fmt_rows.append(".1E")
            elif minexp < 0:
                annot_fmt_rows.append(f".{-minexp + 1}f")
            elif minexp in [0, 1]:
                annot_fmt_rows.append(".1f")
            else:
                annot_fmt_rows.append(".0f")
    if isinstance(annot_fmt_exceed, list):
        exceed_val, exceed_fmt = annot_fmt_exceed
    else:
        exceed_val, exceed_fmt = None, None

    for i, row in enumerate(annot):
        rowfmt = annot_fmt_rows[i]
        if rowfmt == "":
            row_fmt = [""] * len(row)
        else:
            row_fmt = []
            for val in row:
                if np.isnan(val):
                    valstr = ""
                else:
                    if exceed_val is not None and val > exceed_val:
                        valstr = format(val, exceed_fmt)
                    else:
                        valstr = format(val, rowfmt)

                row_fmt.append(valstr)
        _annot.append(row_fmt)
    annot = np.asarray(_annot)
    return annot, annot_fmt_rows


def df_to_heatmap(
    df,
    cmap=None,
    center=None,
    low=0.3,
    high=0.3,
    vmin=None,
    vmax=None,
    color_rowwise=False,
    normalise_rows=False,
    normalise_rows_how=None,
    normalise_rows_col=None,
    norm_ref=None,
    sub_norm_before_div=True,
    annot=True,
    num_digits=None,
    ax=None,
    figsize=(12, 12),
    cbar=False,
    cbar_label=None,
    cbar_labelsize=None,
    xticklabels=None,
    xtick_rot=45,
    yticklabels=None,
    ytick_rot=45,
    xlabel=None,
    ylabel=None,
    title=None,
    labelsize=12,
    annot_fontsize=None,
    annot_fmt_rowwise=False,
    annot_fmt_exceed=None,
    annot_fmt_rows=None,  # explicit formatting strings for rows
    cbar_ax=None,
    cbar_kws=None,
    **kwargs,
):
    """Plot a pandas dataframe as heatmap

    Parameters
    ----------
    df : DataFrame
        table data
    cmap : str, optional
        string specifying colormap to be used
    center : float, optional
        value that is mapped to center colour of colormap (e.g. 0)
    low : float, optional
        Extends lower range of the table values so that when mapped to  the
        colormap, it’s entire range isn’t used. E.g. 0.3 roughly corresponds
        to colormap crop of 30% at the lower end.
    high : float, optional
        Extends upper range of the table values so that when mapped to the
        colormap, it’s entire range isn’t used. E.g. 0.3 roughly corresponds
        to colormap crop of 30% at the upper end.
    vmin : float, optional
        lower end of value range to be plotted. If specified, input arg `low`
        will be ignored.
    vmax : float, optional
        upper end of value range to be plotted. If specified, input arg `low`
        will be ignored.
    color_rowwise : bool, optional
        if True, the color mapping is applied row by row, else, for the whole
        table. Defaults to False.
    normalise_rows : bool, optional
        if True, the table is normalised in a rowwise manner either using the
        mean value in each row (if argument ``normalise_rows_col`` is
        unspecified) or using the value in a specified column. Defaults to
        False.
    normalise_rows_how : str, optional
        aggregation string for row normalisation. Choose from mean or median.
        Only relevant if input arg ``normalise_rows==True``.
    normalise_rows_col : int, optional
        if provided and if arg. ``normalise_rows==True``, then the
        corresponding table column is used for normalisation rather than
        the mean value of each row.
    norm_ref : float or ndarray, optional
        reference value(s) used for rowwise normalisation. Only relevant if
        normalise_rows is True. If specified, normalise_rows_how and
        normalise_rows_col will be ignored.
    sub_norm_before_div : bool, optional
        if True, the rowwise normilisation is applied by subtracting the
        normalisation value for each row before dividing by it. This can be
        useful to visualise positive or negative departures from the mean or
        median.
    annot : bool or list or ndarray, optional
        if True, the table values are printed into the heatmap. Defaults to
        True, in which case the values are computed based on the table content.
        If list or ndarray, the shape needs to be the same as input table shape
        (no of rows and cols), in which case the values of that 2D frame are
        used.
    num_digits : int, optional
        number of digits printed in heatmap annotation.
    ax : axes, optional
        matplotlib axes instance used for plotting, if None, an axes will be
        created
    figsize : tuple, optional
        size of figure for plot
    cbar : bool, optional
        if True, a colorbar is included
    cbar_label : str, optional
        label of colorbar (if colorbar is included, see cbar option)
    cbar_labelsize : int, optional
        size of colorbar label
    xticklabels : list, optional
        List of x axis labels.
    xtick_rot : int, optional
        rotation of x axis labels, defaults to 45 degrees.
    yticklabels : list, optional
        List of string labels.
    ytick_rot : int, optional
        rotation of y axis labels, defaults to 45 degrees.
    xlabel : str, optional
        x axis label
    ylabel : str, optional
        y axis label
    title : str, optional
        title of heatmap
    labelsize : int, optional
        fontsize of labels, default to 12
    annot_fontsize : int, optional
        fontsize of annotated text.
    annot_fmt_rowwise : bool
        rowwise formatting of annotation values, based on row value ranges.
        Defaults to False.
    annot_fmt_exceed : list, optional
        how to format annotated values that exceed a certain threshold.
        The list contains 2 entries, 1. the threshold values, 2. how values
        exceeding this thrshold should be formatted. This parameter is only
        considered if `annot_fmt_rowwise` is True. See also
        :func:`_format_annot_heatmap`.
    annot_fmt_rows : list
        annotation formatting strings for each row of the input table. This
        parameter is only considered if `annot_fmt_rowwise` is True. See also
        :func:`_format_annot_heatmap`.
    cbar_ax : Axes, optional
        axes instance for colorbar, parsed to :func:`seaborn.heatmap`.
    cbar_kws : dict, optional
        keywords for colorbar formatting, , parsed to :func:`seaborn.heatmap`.
    **kwargs
        further keyword args parsed to :func:`seaborn.heatmap`

    Raises
    ------
    ValueError
        if input `annot` is list or ndarray and has a different shape than
        the input `df`.

    Returns
    -------
    Axes
        plot axes instance
    list or None
        annotation information for rows

    """
    warnings.warn(
        "matplotlib based plotting is no longer directly supported. This function may be removed in future versions.",
        DeprecationWarning,
    )

    if cmap is None:
        cmap = "bwr"
    if cbar_label is None:
        cbar_label = ""
    if normalise_rows_how is None:
        normalise_rows_how = "median"

    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=figsize)
    if cbar_kws is None:
        cbar_kws = {}
    if annot_fontsize is None:
        annot_fontsize = labelsize - 4
    if "annot_kws" not in kwargs:
        kwargs["annot_kws"] = {}
    kwargs["annot_kws"]["size"] = annot_fontsize
    df_hm = df
    if normalise_rows:
        if norm_ref is None:
            if normalise_rows_col is not None:
                if isinstance(normalise_rows_col, str):
                    try:
                        normalise_rows_col = df.columns.to_list().index(normalise_rows_col)
                    except ValueError:
                        raise ValueError(f"Failed to localise column {normalise_rows_col}")
                norm_ref = df.values[:, normalise_rows_col]
            else:
                if normalise_rows_how == "mean":
                    norm_ref = df.mean(axis=1)
                elif normalise_rows_how == "median":
                    norm_ref = df.median(axis=1)
                else:
                    raise ValueError(
                        f"Invalid input for normalise_rows_how ({normalise_rows_how}). "
                        f"Choose mean or median"
                    )

        if sub_norm_before_div:
            df_hm = df.subtract(norm_ref, axis=0).div(norm_ref, axis=0)
        else:
            df_hm = df.div(norm_ref, axis=0)
        cbar_kws["format"] = FuncFormatter(lambda x, pos: f"{x:.0%}")

    if color_rowwise:
        df_hm = df_hm.div(abs(df_hm).max(axis=1), axis=0)

    cbar_kws["label"] = cbar_label

    if "norm" in kwargs:
        norm = kwargs["norm"]
        vmin = norm.boundaries[0]
        vmax = norm.boundaries[-1]
    else:
        if vmin is None:
            vmin = df_hm.min().min()  # * (1 - low)
        if vmax is None:
            vmax = df_hm.max().max()  # * (1 + high)
        vmin -= abs(vmin) * low
        vmax += abs(vmax) * high

    if annot is True:
        annot = df.values
    elif isinstance(annot, list):
        annot = np.asarray(annot)
    else:
        fmt = ""

    if isinstance(annot, np.ndarray):
        if not annot.shape == df.values.shape:
            raise ValueError(
                "Invalid input for annot: needs to have same shape as input dataframe"
            )
        elif np.any([isinstance(x, str) for x in annot.flatten()]):
            fmt = ""
        elif annot_fmt_rowwise:
            annot, annot_fmt_rows = _format_annot_heatmap(annot, annot_fmt_rows, annot_fmt_exceed)
            fmt = ""

        elif num_digits is None or num_digits > 5:
            fmt = ".4g"
        else:
            fmt = f".{num_digits}f"

    ax = heatmap(
        df_hm,
        cmap=cmap,
        center=center,
        annot=annot,
        ax=ax,
        # changes this from df_hm to df because the annotation and colorbar didn't work.
        cbar=cbar,
        cbar_ax=cbar_ax,
        cbar_kws=cbar_kws,
        fmt=fmt,
        vmin=vmin,
        vmax=vmax,
        xticklabels=True,
        yticklabels=True,
        **kwargs,
    )

    ax.figure.axes[-1].yaxis.label.set_size(labelsize)
    if title is not None:
        ax.set_title(title, fontsize=labelsize + 2)

    if yticklabels is None:
        yticklabels = ax.get_yticklabels()

    ax.set_yticklabels(yticklabels, rotation=ytick_rot, fontsize=labelsize - 2)

    if xticklabels is None:
        xticklabels = ax.get_xticklabels()

    ax.set_xticklabels(xticklabels, rotation=xtick_rot, ha="right", fontsize=labelsize - 2)

    if xlabel is None:
        xlabel = ""

    ax.set_xlabel(xlabel, fontsize=labelsize)

    if ylabel is None:
        ylabel = ""
    ax.set_ylabel(ylabel, fontsize=labelsize)
    if cbar_labelsize is not None:
        ax.figure.axes[-1].yaxis.label.set_size(cbar_labelsize)

    return ax, annot_fmt_rows



import iris
import numpy as np
import pandas as pd
import pytest
from cf_units import Unit

from pyaerocom import GriddedData, const, helpers
from pyaerocom.colocation.colocated_data import ColocatedData
from pyaerocom.colocation.colocation_utils import (
    _colocate_site_data_helper,
    _colocate_site_data_helper_timecol,
    _regrid_gridded,
    colocate_gridded_gridded,
    colocate_gridded_ungridded,
)
from pyaerocom.config import ALL_REGION_NAME
from pyaerocom.exceptions import UnresolvableTimeDefinitionError
from pyaerocom.io.mscw_ctm.reader import ReadMscwCtm
from tests.conftest import TEST_RTOL, need_iris_32
from tests.fixtures.stations import create_fake_station_data


def test__regrid_gridded(data_tm5):
    one_way = _regrid_gridded(data_tm5, "areaweighted", 5)
    another_way = _regrid_gridded(data_tm5, "areaweighted", dict(lon_res_deg=5, lat_res_deg=5))

    assert one_way.shape == another_way.shape


S1 = create_fake_station_data(
    "concpm10",
    {"concpm10": {"units": "ug m-3"}},
    10,
    "2010-01-01",
    "2010-12-31",
    "d",
    {"ts_type": "daily"},
)

S1["concpm10"][10:20] = np.nan

S2 = create_fake_station_data(
    "concpm10",
    {"concpm10": {"units": "ug m-3"}},
    10,
    "2010-01-01",
    "2010-12-31",
    "d",
    {"ts_type": "daily"},
)

S3 = create_fake_station_data(
    "concpm10",
    {"concpm10": {"units": "ug m-3"}},
    10,
    "2010-01-01",
    "2010-12-31",
    "13d",
    {"ts_type": "13daily"},
)
S3["concpm10"][1] = np.nan
S3["concpm10"][3] = np.nan

S4 = create_fake_station_data(
    "concpm10",
    {"concpm10": {"units": "ug m-3"}},
    10,
    "2010-01-03",
    "2011-12-31",
    "d",
    {"ts_type": "daily"},
)

S4["concpm10"][0:5] = range(5)


@pytest.mark.parametrize(
    "stat_data,stat_data_ref,var,var_ref,ts_type,resample_how,min_num_obs, use_climatology_ref,num_valid",
    [
        (S4, S3, "concpm10", "concpm10", "monthly", "mean", {"monthly": {"daily": 25}}, False, 10),
        (S3, S4, "concpm10", "concpm10", "monthly", "mean", {"monthly": {"daily": 25}}, False, 24),
        (S1, S2, "concpm10", "concpm10", "monthly", "mean", 25, False, 12),
        (S2, S1, "concpm10", "concpm10", "monthly", "mean", 25, False, 11),
    ],
)
def test__colocate_site_data_helper_timecol(
    stat_data,
    stat_data_ref,
    var,
    var_ref,
    ts_type,
    resample_how,
    min_num_obs,
    use_climatology_ref,
    num_valid,
):
    result = _colocate_site_data_helper_timecol(
        stat_data,
        stat_data_ref,
        var,
        var_ref,
        ts_type,
        resample_how,
        min_num_obs,
        use_climatology_ref,
    )

    assert isinstance(result, pd.DataFrame)
    assert result.data.isnull().sum() == result.ref.isnull().sum()
    assert len(result) - result.data.isnull().sum() == num_valid


def test__colocate_site_data_helper(aeronetsunv3lev2_subset):
    var = "od550aer"
    stat1 = aeronetsunv3lev2_subset.to_station_data(3, var)
    stat2 = aeronetsunv3lev2_subset.to_station_data(4, var)
    df = _colocate_site_data_helper(stat1, stat2, var, var, "daily", None, None, False)

    assert isinstance(df, pd.DataFrame)
    assert len(df) == 9483
    assert df["data"].mean() == pytest.approx(0.31171085422102346, rel=TEST_RTOL)
    assert df["ref"].mean() == pytest.approx(0.07752743643132792, rel=TEST_RTOL)


def test_colocate_gridded_ungridded_new_var(data_tm5, aeronetsunv3lev2_subset):
    data = data_tm5.copy()
    data.var_name = "od550bc"
    coldata = colocate_gridded_ungridded(data, aeronetsunv3lev2_subset, var_ref="od550aer")

    assert coldata.metadata["var_name"] == ["od550aer", "od550bc"]


@pytest.mark.parametrize(
    "addargs,ts_type,shape,obsmean,modmean",
    [
        (
            dict(
                filter_name=f"{ALL_REGION_NAME}-noMOUNTAINS",
                min_num_obs=const.OBS_MIN_NUM_RESAMPLE,
            ),
            "monthly",
            (2, 12, 8),
            0.315930,
            0.275671,
        ),
        (
            dict(filter_name=f"{ALL_REGION_NAME}-noMOUNTAINS"),
            "monthly",
            (2, 12, 8),
            0.316924,
            0.275671,
        ),
        (
            dict(
                filter_name=f"{ALL_REGION_NAME}-wMOUNTAINS", min_num_obs=const.OBS_MIN_NUM_RESAMPLE
            ),
            "monthly",
            (2, 12, 11),
            0.269707,
            0.243861,
        ),
        (
            dict(
                filter_name=f"{ALL_REGION_NAME}-noMOUNTAINS",
                use_climatology_ref=True,
                min_num_obs=const.OBS_MIN_NUM_RESAMPLE,
            ),
            "monthly",
            (2, 12, 13),
            0.302636,
            0.234147,
        ),
        pytest.param(
            dict(
                filter_name=f"{ALL_REGION_NAME}-noMOUNTAINS",
                regrid_res_deg=30,
                min_num_obs=const.OBS_MIN_NUM_RESAMPLE,
            ),
            "monthly",
            (2, 12, 8),
            0.31593,
            # 0.1797,
            0.169897,
            marks=[need_iris_32],
        ),
        (
            dict(filter_name=f"{ALL_REGION_NAME}-noMOUNTAINS", ts_type="yearly"),
            "yearly",
            (2, 1, 8),
            0.417676,
            0.275671,
        ),
    ],
)
def test_colocate_gridded_ungridded(
    data_tm5, aeronetsunv3lev2_subset, addargs, ts_type, shape, obsmean, modmean
):
    coldata = colocate_gridded_ungridded(data_tm5, aeronetsunv3lev2_subset, **addargs)

    assert isinstance(coldata, ColocatedData)
    assert coldata.ts_type == ts_type
    assert coldata.shape == shape

    assert np.nanmean(coldata.data.data[0]) == pytest.approx(obsmean, rel=TEST_RTOL)
    assert np.nanmean(coldata.data.data[1]) == pytest.approx(modmean, rel=TEST_RTOL)


def test_colocate_gridded_ungridded_nonglobal(aeronetsunv3lev2_subset):
    times = [1, 2]
    time_unit = Unit("days since 2010-1-1 0:0:0")
    cubes = iris.cube.CubeList()

    for time in times:
        time_coord = iris.coords.DimCoord(time, units=time_unit, standard_name="time")
        cube = helpers.make_dummy_cube_latlon(
            lat_res_deg=1, lon_res_deg=1, lat_range=[30.05, 81.95], lon_range=[-29.5, 89.95]
        )
        cube.add_aux_coord(time_coord)
        cubes.append(cube)
    time_cube = cubes.merge_cube()
    gridded = GriddedData(time_cube)
    gridded.var_name = "od550aer"
    gridded.units = Unit("1")

    coldata = colocate_gridded_ungridded(gridded, aeronetsunv3lev2_subset, colocate_time=False)
    assert isinstance(coldata, ColocatedData)
    assert coldata.shape == (2, 2, 2)


def test_colocate_gridded_gridded_same_new_var(data_tm5):
    data = data_tm5.copy()
    data.var_name = "Blaaa"
    coldata = colocate_gridded_gridded(data, data_tm5)

    assert coldata.metadata["var_name"] == ["od550aer", "Blaaa"]


def test_colocate_gridded_gridded_same(data_tm5):
    coldata = colocate_gridded_gridded(data_tm5, data_tm5)

    assert isinstance(coldata, ColocatedData)
    stats = coldata.calc_statistics()
    # check mean value
    assert stats["data_mean"] == pytest.approx(0.09825691)
    # check that mean value is same as in input GriddedData object
    assert stats["data_mean"] == pytest.approx(data_tm5.mean(areaweighted=False))
    assert stats["refdata_mean"] == stats["data_mean"]
    assert stats["nmb"] == 0
    assert stats["mnmb"] == 0
    assert stats["R"] == pytest.approx(1, rel=1e-6)
    assert stats["R_spearman"] == pytest.approx(1, rel=1e-6)


@pytest.mark.xfail(raises=UnresolvableTimeDefinitionError)
def test_read_emep_colocate_emep_tm5(data_tm5, path_emep):
    reader = ReadMscwCtm(data_dir=path_emep["data_dir"])
    data_emep = reader.read_var("concpm10", ts_type="monthly")

    # Change units and year to match TM5 data
    data_emep.change_base_year(2010)
    data_emep.units = "1"

    col = colocate_gridded_gridded(data_emep, data_tm5)
    assert isinstance(col, ColocatedData)



import logging
from pathlib import Path
from traceback import format_exc
from typing import Literal

from pydantic import (
    BaseModel,
    ConfigDict,
    field_validator,
    model_validator,
)

from pyaerocom import const
from pyaerocom._lowlevel_helpers import LayerLimits
from pyaerocom.exceptions import InitialisationError

logger = logging.getLogger(__name__)


SUPPORTED_VERT_CODES: tuple[
    str,
    str,
    str,
] = (
    "Column",
    "Profile",
    "Surface",
)

ALT_NAMES_VERT_CODES: dict = dict(ModelLevel="Profile")


SUPPORTED_VERT_LOCS: tuple[str, str, str] = (
    "ground",
    "space",
    "airborne",
)


class ObsEntry(BaseModel):
    """Observation configuration for evaluation (BaseModel)

    Note
    ----
    Only :attr:`obs_id` and `obs_vars` are mandatory, the rest are optional.

    Attributes
    ----------
    obs_id : str
        ID of observation network in AeroCom database
        (e.g. 'AeronetSunV3Lev2.daily')
        Note that this can also be a custom supplied obs_id if and only if bs_aux_requires is provided
    obs_vars : tuple[str, ...]
        tuple of pyaerocom variable names that are supposed to be analysed
        (e.g. ('od550aer', 'ang4487aer'))
    obs_ts_type_read : :obj:`str` or :obj:`dict`, optional
        may be specified to explicitly define the reading frequency of the
        observation data (so far, this does only apply to gridded obsdata such
        as satellites). For ungridded reading, the frequency may be specified
        via :attr:`obs_id`, where applicable (e.g. AeronetSunV3Lev2.daily).
        Can be specified variable specific in form of dictionary.
    obs_vert_type : str, optional
        Aerocom vertical code encoded in the model filenames (only AeroCom 3
        and later).
    obs_aux_requires : dict, optional
        information about required datasets / variables for auxiliary
        variables.
    instr_vert_loc : str, optional
        vertical location code of observation instrument. This is used in
        the aeroval interface for separating different categories of measurements
        such as "ground", "space" or "airborne".
    is_superobs : bool
        if True, this observation is a combination of several others which all
        have to have their own obs config entry.
    only_superobs : bool
        this indicates whether this configuration is only to be used as part
        of a superobs network, and not individually.
    read_opts_ungridded : :obj:`dict`, optional
        dictionary that specifies reading constraints for ungridded reading
        (c.g. :class:`pyaerocom.io.ReadUngridded`).
    only_json : bool
        Only to be set if the obs entry already has colocated data files which were
        preprocessed outside of pyaerocom. Setting to True will skip the colcoation
        and just create the JSON output.
    coldata_dir : str
        Only to be set if the obs entry already has colocated data files which were
        preprocessed outside of pyaerocom. This is the directory in which the
        colocated data files are located.

    """

    ##   Pydantic ConfigDict
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="allow",
        validate_assignment=True,
    )

    ######################
    ## Required attributes
    ######################
    obs_vars: str | tuple[str, ...]
    obs_id: str | tuple[str, ...]
    ######################
    ## Optional attributes
    ######################
    obs_name: str | None = None  # not expected to be set directly, rather set by ObsCollection
    obs_ts_type_read: str | dict | None = None
    obs_vert_type: Literal["Column", "Profile", "Surface", "ModelLevel"] = "Surface"
    obs_aux_requires: dict[str, dict] = {}
    instr_vert_loc: str | None = None
    is_superobs: bool = False
    only_superobs: bool = False
    colocation_layer_limts: tuple[LayerLimits, ...] | None = None
    profile_layer_limits: tuple[LayerLimits, ...] | None = None
    web_interface_name: str | None = None
    diurnal_only: bool = False
    obs_type: str | None = None

    read_opts_ungridded: dict = {}
    # attributes for reading colocated data files made outside of pyaerocom
    only_json: bool = False
    coldata_dir: str | Path | None = (
        None  # TODO: Would like this to be a Path but need to see if it will cause issues down the line
    )

    #############
    ## Validators
    #############
    @field_validator("obs_vars")
    @classmethod
    def validate_obs_vars(cls, v):
        if isinstance(v, str):
            return (v,)
        return v

    @field_validator("instr_vert_loc")
    @classmethod
    def validate_instr_vert_loc(cls, v):
        if isinstance(v, str) and v not in SUPPORTED_VERT_LOCS:
            raise AttributeError(
                f"Invalid value for instr_vert_loc: {v} for {cls.obs_id}. "
                f"Please choose from: {SUPPORTED_VERT_LOCS}"
            )

    @field_validator("obs_vert_type")
    @classmethod
    def check_obs_vert_type(cls, ovt):
        """Check if obs_vert_type string is valid alias
        Parameters
        ----------
        ovt : str
            obs_vert_type string
        Returns
        -------
        str
            valid obs_vert_type
        Raises
        ------
        ValueError
            if `ovt` is invalid
        """
        if ovt in SUPPORTED_VERT_CODES:
            return ovt
        if ovt in ALT_NAMES_VERT_CODES:
            logger.warning(
                f"Please use {ALT_NAMES_VERT_CODES[ovt]} for obs_vert_code and not {ovt}"
            )
            ovt = ALT_NAMES_VERT_CODES[ovt]
            return ovt
        valid = SUPPORTED_VERT_CODES + list(ALT_NAMES_VERT_CODES)
        raise ValueError(
            f"Invalid value for obs_vert_type: {ovt}. " f"Supported codes are {valid}."
        )

    @model_validator(mode="after")
    def check_cfg(self):
        if not self.is_superobs and not isinstance(self.obs_id, str | tuple | dict):
            raise ValueError(
                f"Invalid value for obs_id: {self.obs_id}. Need str, tuple, or dict "
                f"or specification of ids and variables via obs_compute_post"
            )
        self.check_add_obs()
        return self

    ##########
    ## Methods
    ##########

    def check_add_obs(self):
        """Check if this dataset is an auxiliary post dataset"""
        if len(self.obs_aux_requires) > 0:
            if not self.obs_type == "ungridded":
                raise NotImplementedError(
                    f"Cannot initialise auxiliary setup for {self.obs_id}. "
                    f"Aux obs reading is so far only possible for ungridded observations."
                )
            if self.obs_id not in const.OBS_IDS_UNGRIDDED:
                try:
                    const.add_ungridded_post_dataset(**self.model_dump())
                except Exception:
                    raise InitialisationError(
                        f"Cannot initialise auxiliary reading setup for {self.obs_id}. "
                        f"Reason:\n{format_exc()}"
                    )

    def get_all_vars(self) -> tuple[str, ...]:
        """
        Get a tuple of all variables associated with this entry

        Returns
        -------
        tuple[str, ...]
        """
        return self.obs_vars

    def has_var(self, var_name):
        """
        Check if input variable is defined in entry

        Returns
        -------
        bool
            True if entry has variable available, else False
        """
        return True if var_name in self.get_all_vars() else False

    def get_vert_code(self, var):
        """Get vertical code name for obs / var combination"""
        vc = self.obs_vert_type
        if isinstance(vc, str):
            val = vc
        elif isinstance(vc, dict) and var in vc:
            val = vc[var]
        else:
            raise ValueError(f"invalid value for obs_vert_type: {vc}")
        if val not in SUPPORTED_VERT_CODES:
            raise ValueError(
                f"invalid value for obs_vert_type: {val}. Choose from " f"{SUPPORTED_VERT_CODES}."
            )
        return val



from __future__ import annotations

import numpy as np
from pytest import mark, param

from pyaerocom.aeroval.coldatatojson_helpers import (  # _add_heatmap_entry_json,; _write_diurnal_week_stationdata_json,; _write_site_data,; _write_stationdata_json,; get_stationfile_name,
    _init_stats_dummy,
    _prepare_aerocom_regions_json,
    _prepare_country_regions,
    _prepare_default_regions_json,
    _prepare_htap_regions_json,
    _prepare_regions_json_helper,
)
from pyaerocom.region import get_all_default_region_ids
from pyaerocom.region_defs import (
    HTAP_REGIONS,
    HTAP_REGIONS_DEFAULT,
    OLD_AEROCOM_REGIONS,
    OTHER_REGIONS,
)


def test__init_stats_dummy():
    dummy = _init_stats_dummy()
    assert dummy
    assert all(np.isnan(value) for value in dummy.values())


@mark.parametrize(
    "region_ids",
    [
        param(OLD_AEROCOM_REGIONS, id="old AeroCom"),
        param(HTAP_REGIONS_DEFAULT, id="HTAP default"),
        param(HTAP_REGIONS, id="HTAP"),
        param(OTHER_REGIONS, id="other"),
    ],
)
def test__prepare_regions_json_helper(region_ids: list[str]):
    helper = _prepare_regions_json_helper(region_ids)
    assert len(helper) == 2
    assert len(helper[0]) == len(helper[1]) == len(region_ids)

    for region_name, borders in helper[0].items():
        assert region_name in helper[1]
        assert type(borders) is dict
        assert "minLat" in borders and -90 <= borders["minLat"] <= 90
        assert "maxLat" in borders and -90 <= borders["maxLat"] <= 90
        assert "minLon" in borders and -180 <= borders["minLon"] <= 180
        assert "maxLon" in borders and -180 <= borders["maxLon"] <= 180
        assert borders["minLat"] < borders["maxLat"]

    for region_name, region in helper[1].items():
        assert region_name in helper[0]
        assert region.region_id in region_ids
        assert region.name == region_name


def test__prepare_default_regions_json():
    borders, regions = _prepare_default_regions_json()
    region_ids = get_all_default_region_ids()
    assert bool(borders) and bool(regions) and bool(region_ids)
    assert list(borders) == list(regions)
    assert [reg.region_id for reg in regions.values()] == region_ids


def test__prepare_aerocom_regions_json():
    borders, regions = _prepare_aerocom_regions_json()
    region_ids = OLD_AEROCOM_REGIONS
    assert bool(borders) and bool(regions) and bool(region_ids)
    assert list(borders) == list(regions)
    assert [reg.region_id for reg in regions.values()] == region_ids


def test__prepare_htap_regions_json():
    borders, regions = _prepare_htap_regions_json()
    region_ids = HTAP_REGIONS_DEFAULT
    assert bool(borders) and bool(regions) and bool(region_ids)
    assert list(borders) == list(regions)
    assert [reg.region_id for reg in regions.values()] == region_ids


@mark.parametrize(
    "region_ids",
    [
        param(OLD_AEROCOM_REGIONS, id="old AeroCom"),
        param(HTAP_REGIONS_DEFAULT, id="HTAP default"),
        param(HTAP_REGIONS, id="HTAP"),
        param(OTHER_REGIONS, id="other"),
    ],
)
def test__prepare_country_regions(region_ids: list[str]):
    regions = _prepare_country_regions(region_ids)
    assert type(regions) is dict
    assert bool(regions)
    assert [reg.region_id for reg in regions.values()] == region_ids
